import os
import unittest

from orangewidget.tests.base import WidgetTest

from Orange.data import Table

from orangecontrib.bioinformatics.widgets.OWGenes import OWGenes
from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID,
    GENE_ID_COLUMN,
    GENE_AS_ATTRIBUTE_NAME,
)


class TestOWGenes(WidgetTest):
    def setUp(self):
        self._path = os.path.join(os.path.dirname(__file__), 'data')
        self.widget = self.create_widget(OWGenes)
        self.markers_file = 'genes.tab'

        file_name = os.path.join(self._path, self.markers_file)

        # gene names are stored in "Name" column.
        # we expect widget to detect this on its own.
        self.markers_column = 'Name'
        self.target_db = 'Entrez ID'

        # get input data
        self.send_signal(self.widget.Inputs.data_table, Table(file_name))
        self.assertIsNotNone(self.widget.input_data)

        # check if settings were set correctly
        self.assertEqual(self.markers_column, self.widget.selected_gene_col.name)
        self.assertFalse(self.widget.use_attr_names)

        # check if genes were extracted from the table
        self.assertIsNotNone(self.widget.input_genes)
        genes = self.widget.input_data.get_column(self.widget.selected_gene_col)
        self.assertEqual(len(self.widget.input_genes), len(genes))

        self.wait_until_finished()
        # wait for gene matcher to finish
        # self.widget.threadpool.waitForDone()
        # self.process_events()

        # gene matcher should map all of the marker names into corresponding Entrez ID
        self.gene_match_result = self.widget.gene_matcher.get_known_genes()
        self.assertEqual(len(genes), len(self.gene_match_result))

    def test_output_data(self):
        # get output data
        out_data = self.get_output(self.widget.Outputs.data_table)

        # check if 'target_db' column exists. Target database is Entrez
        self.assertTrue(self.target_db in out_data.domain)

        # test if genes in the output data is the same as known genes from gene matcher.
        ids = out_data.get_column(self.target_db)
        self.assertTrue((ids == [str(g.gene_id) for g in self.gene_match_result]).all())

        # test if table on the output is properly annotated
        table_annotations = out_data.attributes
        self.assertTrue(len(table_annotations) > 0)
        self.assertTrue(GENE_ID_COLUMN in table_annotations)
        self.assertTrue(TAX_ID in table_annotations)
        self.assertTrue(GENE_AS_ATTRIBUTE_NAME in table_annotations)
        self.assertFalse(table_annotations[GENE_AS_ATTRIBUTE_NAME])
        self.assertTrue(table_annotations[GENE_ID_COLUMN] == self.target_db)
        self.assertTrue(table_annotations[TAX_ID] == self.widget.tax_id)

    def test_filtered_output_data(self):
        self.widget.search_pattern = 'tubulin gamma'
        self.widget._apply_filter()

        # get output data
        out_data = self.get_output(self.widget.Outputs.data_table)

        # check if 'target_db' column exists. Target database is Entrez
        self.assertTrue(self.target_db in out_data.domain)

        # we expect filtered data table on the output
        ids = out_data.get_column(self.target_db)
        self.assertTrue(len(ids) == 2)

        # test if table on the output is properly annotated
        table_annotations = out_data.attributes
        self.assertTrue(len(table_annotations) > 0)
        self.assertTrue(GENE_ID_COLUMN in table_annotations)
        self.assertTrue(TAX_ID in table_annotations)
        self.assertTrue(GENE_AS_ATTRIBUTE_NAME in table_annotations)
        self.assertFalse(table_annotations[GENE_AS_ATTRIBUTE_NAME])
        self.assertTrue(table_annotations[GENE_ID_COLUMN] == self.target_db)
        self.assertTrue(table_annotations[TAX_ID] == self.widget.tax_id)


if __name__ == '__main__':
    unittest.main()



""" GeneSets """
from typing import Set, List, Tuple, Optional
from urllib.parse import urlparse

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QTableView, QHBoxLayout

from Orange.data import Table, Domain
from Orange.data import filter as table_filter
from Orange.widgets.gui import (
    LinkRole,
    spin,
    vBox,
    deferred,
    lineEdit,
    widgetBox,
    auto_commit,
)
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.settings import Setting, SettingProvider
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.itemmodels import PyTableModel

from orangecontrib.bioinformatics.geneset import GeneSet, GeneSets
from orangecontrib.bioinformatics.widgets.utils.gui import FilterProxyModel
from orangecontrib.bioinformatics.widgets.components import GeneSetSelection
from orangecontrib.bioinformatics.widgets.utils.data import (
    TableAnnotation,
    check_table_annotation,
)
from orangecontrib.bioinformatics.widgets.utils.itemmodels import TableModel
from orangecontrib.bioinformatics.widgets.utils.itemdelegates import (
    LinkStyledItemDelegate,
)


def run(
    gene_sets: GeneSets,
    selected_gene_sets: List[Tuple[str, ...]],
    genes,
    state: TaskState,
) -> Tuple[List[int], List[GeneSet]]:
    step, steps = 0, len(gene_sets)

    filtered_gene_sets = []
    mapped_genes = []

    for gene_set in sorted(gene_sets):
        step += 1
        if step % (steps / 10) == 0:
            state.set_progress_value(100 * step / steps)

        if gene_set.hierarchy not in selected_gene_sets:
            continue

        if state.is_interruption_requested():
            return mapped_genes, filtered_gene_sets

        mapped_genes.append(gene_set.genes & genes)
        filtered_gene_sets.append(gene_set)

    return filtered_gene_sets, mapped_genes


class GeneSetsModel(TableModel):
    (
        mapped_genes,
        genes_in_set,
        collection,
        gs_name,
    ) = range(4)

    def __init__(self, gs: List[GeneSet], matching_genes: List[str]):
        def gene_set_url(row):
            return (
                gs[row].link if gs[row].link and urlparse(gs[row].link).scheme else None
            )

        columns = [
            TableModel.Column(
                {Qt.DisplayRole: 'Mapped Genes'},
                {
                    Qt.DisplayRole: lambda row: len(matching_genes[row]),
                    Qt.UserRole: lambda row: matching_genes[row],
                },
            ),
            TableModel.Column(
                {Qt.DisplayRole: 'Genes In Set'},
                {
                    Qt.DisplayRole: lambda row: len(gs[row].genes),
                    Qt.UserRole: lambda row: gs[row],
                },
            ),
            TableModel.Column(
                {Qt.DisplayRole: 'Collection'},
                {
                    Qt.DisplayRole: lambda row: ', '.join(gs[row].hierarchy),
                },
            ),
            TableModel.Column(
                {Qt.DisplayRole: 'Name'},
                {
                    Qt.DisplayRole: lambda row: gs[row].name,
                    LinkRole: gene_set_url,
                    Qt.ToolTipRole: gene_set_url,
                },
            ),
        ]

        super().__init__(len(gs), columns)

    def sortColumnData(self, column):
        return super().sortColumnData(column)


class OWGeneSets(OWWidget, ConcurrentWidgetMixin):
    name = 'Gene Sets'
    description = ""
    icon = 'icons/OWGeneSets.svg'
    priority = 80
    want_main_area = True

    organism = Setting(None, schema_only=True)
    stored_gene_sets_selection = Setting([], schema_only=True)
    selected_rows = Setting([], schema_only=True)

    min_count: int
    min_count = Setting(5)

    use_min_count: bool
    use_min_count = Setting(False)

    auto_commit: bool
    auto_commit = Setting(True)

    search_pattern: str
    search_pattern = Setting('')

    # component settings
    gs_selection_component: SettingProvider = SettingProvider(GeneSetSelection)

    class Inputs:
        data = Input('Data', Table)
        custom_gene_sets = Input('Custom Gene Sets', Table)

    class Outputs:
        gene_sets = Output('Gene Sets', GeneSets)
        reduced_data = Output('Reduced Data', Table)

    class Warning(OWWidget.Warning):
        all_sets_filtered = Msg('All sets were filtered out.')

    class Error(OWWidget.Error):
        organism_mismatch = Msg(
            'Organism in input data and custom gene sets does not match'
        )
        cant_reach_host = Msg('Host orange.biolab.si is unreachable.')
        cant_load_organisms = Msg(
            'No available organisms, please check your connection.'
        )
        custom_gene_sets_table_format = Msg(
            'Custom gene sets data must have genes represented as rows.'
        )

    def __init__(self):
        super().__init__()
        # OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.input_data: Optional[Table] = None

        # Control area
        box = vBox(self.controlArea, True, margin=0)
        self.gs_selection_component: GeneSetSelection = GeneSetSelection(self, box)
        self.gs_selection_component.selection_changed.connect(
            self._on_selection_changed
        )
        self.gs_selection_component.initialize()

        # Main area
        self.filter_proxy_model = FilterProxyModel()

        self.view = QTableView()
        self.view.setSelectionBehavior(QTableView.SelectRows)
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.setAlternatingRowColors(True)
        self.view.setSortingEnabled(True)
        self.view.verticalHeader().hide()
        self.view.viewport().setMouseTracking(True)
        self.view.setItemDelegateForColumn(
            GeneSetsModel.gs_name, LinkStyledItemDelegate(self.view)
        )
        self.view.setModel(self.filter_proxy_model)

        h_layout = QHBoxLayout()
        h_layout.setSpacing(100)
        h_widget = widgetBox(self.mainArea, orientation=h_layout)

        spin(
            h_widget,
            self,
            'min_count',
            0,
            1000,
            label='Min mapped genes',
            tooltip='Minimum amount of input genes that map to gene set.',
            checked='use_min_count',
            callback=self.filter_view,
            callbackOnReturn=True,
            checkCallback=self.filter_view,
        )

        self.line_edit_filter = lineEdit(h_widget, self, 'search_pattern')
        self.line_edit_filter.setPlaceholderText('Filter gene sets ...')
        self.line_edit_filter.textChanged.connect(self.filter_view)

        self.mainArea.layout().addWidget(self.view)
        self.commit_button = auto_commit(
            self.controlArea, self, 'auto_commit', '&Commit', box=False
        )

    @property
    def gene_as_attr_name(self) -> Optional[bool]:
        if self.input_data:
            return self.input_data.attributes[TableAnnotation.gene_as_attr_name]

    @property
    def gene_location(self) -> Optional[str]:
        if not self.input_data:
            return

        if self.gene_as_attr_name:
            return self.input_data.attributes[TableAnnotation.gene_id_attribute]
        else:
            return self.input_data.attributes[TableAnnotation.gene_id_column]

    @property
    def input_genes(self) -> Set[str]:
        if not self.input_data:
            return set()

        if self.gene_as_attr_name:
            return {
                str(variable.attributes.get(self.gene_location, '?'))
                for variable in self.input_data.domain.attributes
            }
        else:
            return {str(g) for g in self.input_data.get_column(self.gene_location)}

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Tuple):
        self.filter_proxy_model.setSourceModel(GeneSetsModel(*result))

        if self.input_data:
            self.view.sortByColumn(GeneSetsModel.mapped_genes, Qt.DescendingOrder)
            self.view.setColumnHidden(GeneSetsModel.mapped_genes, False)
        else:
            self.view.sortByColumn(GeneSetsModel.genes_in_set, Qt.DescendingOrder)
            self.view.setColumnHidden(GeneSetsModel.mapped_genes, True)

        self.view.resizeColumnsToContents()
        # this is slow
        # self.view.resizeRowsToContents()
        self.view.selectionModel().selectionChanged.connect(self.commit.deferred)
        self.filter_view()

    def on_exception(self, ex):
        # TODO: handle possible exceptions
        raise ex

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def _on_selection_changed(self):
        if hasattr(self, 'input_genes'):
            self.start(
                run,
                self.gs_selection_component.gene_sets,
                self.gs_selection_component.selection,
                self.input_genes,
            )

    @check_table_annotation
    @Inputs.data
    def set_data(self, input_data: Table):
        self.Outputs.gene_sets.send(None)
        self.Outputs.reduced_data.send(None)
        self.input_data = None

        if input_data:
            self.input_data = input_data
            tax_id = self.input_data.attributes[TableAnnotation.tax_id]
            self.gs_selection_component.set_selected_organism_by_tax_id(tax_id)

    @Inputs.custom_gene_sets
    def handle_custom_gene_sets_input(self, custom_data):
        self.Outputs.gene_sets.send(None)
        self.Outputs.reduced_data.send(None)

        if custom_data:
            self.gs_selection_component.initialize_custom_gene_sets(custom_data)
        else:
            self.gs_selection_component.initialize_custom_gene_sets(None)

    @deferred
    def commit(self):
        selection_model = self.view.selectionModel()
        gene_sets = None
        reduced_data = None

        if selection_model:
            selection_gene_set = selection_model.selectedRows(
                GeneSetsModel.genes_in_set
            )
            gene_sets = GeneSets(
                [model_index.data(Qt.UserRole) for model_index in selection_gene_set]
            )

            selection_mapped_genes = selection_model.selectedRows(
                GeneSetsModel.mapped_genes
            )
            self.selected_rows = [
                self.filter_proxy_model.mapToSource(sel).row()
                for sel in selection_mapped_genes
            ]

            if selection_mapped_genes and self.input_genes:
                genes = [
                    model_index.data(Qt.UserRole)
                    for model_index in selection_mapped_genes
                ]
                output_genes = list(set.union(*genes))

                if self.gene_as_attr_name:
                    selected = [
                        column
                        for column in self.input_data.domain.attributes
                        if self.gene_location in column.attributes
                        and str(column.attributes[self.gene_location]) in output_genes
                    ]
                    domain = Domain(
                        selected,
                        self.input_data.domain.class_vars,
                        self.input_data.domain.metas,
                    )
                    reduced_data = self.input_data.from_table(domain, self.input_data)
                else:
                    # create filter from selected column for genes
                    only_known = table_filter.FilterStringList(
                        self.gene_location, output_genes
                    )
                    # apply filter to the data
                    reduced_data = table_filter.Values([only_known])(self.input_data)

        self.Outputs.gene_sets.send(gene_sets)
        self.Outputs.reduced_data.send(reduced_data)

    def create_filters(self):
        search_term: List[str] = self.search_pattern.lower().strip().split()
        filters = []

        if search_term:
            filters.append(
                FilterProxyModel.Filter(
                    GeneSetsModel.gs_name,
                    Qt.DisplayRole,
                    lambda value: all(fs in value.lower() for fs in search_term),
                )
            )

        if self.use_min_count:
            filters.append(
                FilterProxyModel.Filter(
                    GeneSetsModel.mapped_genes,
                    Qt.DisplayRole,
                    lambda value: value >= self.min_count,
                )
            )

        return filters

    def filter_view(self):
        filter_proxy: FilterProxyModel = self.filter_proxy_model
        model: PyTableModel = filter_proxy.sourceModel()

        # apply filtering rules
        filter_proxy.set_filters(self.create_filters())

        if model.rowCount() and not filter_proxy.rowCount():
            self.Warning.all_sets_filtered()
        else:
            self.Warning.clear()

        # this is slow
        # self.view.resizeRowsToContents()

    # def sizeHint(self):
    #     return QSize(800, 600)


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    widget = WidgetPreview(OWGeneSets)
    widget.run()



""" Utils """
import io
import os
import gzip
import json

from Orange.data import Table, Domain, StringVariable, ContinuousVariable

from orangecontrib.bioinformatics.utils import local_cache

#  Support cache with requests_cache module
LOCAL_CACHE_DIR = os.path.join(local_cache, 'resolwe')
try:
    os.makedirs(LOCAL_CACHE_DIR)
except OSError:
    pass

CACHE_BACKEND = 'sqlite'
GENAPI_CACHE = os.path.join(LOCAL_CACHE_DIR, 'GenAPI_cache')
RESOLWEAPI_CACHE = os.path.join(LOCAL_CACHE_DIR, 'ResolweAPI_cache')


def etc_to_table(etc_json, transpose=False):
    """Converts data from Json to :obj:`Orange.data.table`

    Args:
        etc_json (dict): Data in json like format from genesis
        transpose (bool): Transpose table so that genes are in columns.
                          Default is set to False.

    Returns:
        :obj:`Orange.data.Table`
    """

    variables = []
    time_point = 1
    for time in etc_json['etc']['timePoints']:
        var = ContinuousVariable(str(time))
        var.attributes['Time point'] = str(time_point)
        variables.append(var)
        time_point += 1

    meta_attr = StringVariable.make('Gene')
    domain = Domain(variables, metas=[meta_attr])

    table = []
    for row in etc_json['etc']['genes']:
        gene_expression = list(etc_json['etc']['genes'][row])
        gene_expression.append(row)
        table.append(gene_expression)

    orange_table = Table.from_list(domain, table)
    if transpose:
        orange_table = Table.transpose(
            orange_table,
            feature_names_column=meta_attr.name,
            meta_attr_name='Time',
            remove_redundant_inst=True,
        )

        old_domain = orange_table.domain
        new_domain = Domain(
            old_domain.attributes,
            old_domain.class_vars,
            [ContinuousVariable(var.name) for var in old_domain.metas],
        )

        time_as_floats = orange_table.get_column('Time').astype(int).reshape(-1, 1)
        orange_table = orange_table.transform(new_domain)
        orange_table[:, new_domain['Time']] = time_as_floats
    return orange_table


def response_to_json(response):
    if not response.ok:
        response.raise_for_status()

    response_gzipped = io.BytesIO(response.content)
    response_content = io.TextIOWrapper(
        gzip.GzipFile(fileobj=response_gzipped), encoding='utf-8'
    )

    try:
        json_object = json.load(response_content)
    except ValueError as e:
        raise ValueError('Downloaded data is not a valid JSON') from e

    return json_object



""" KEGG Pathway """
import gc
import sys
import webbrowser
from typing import Dict, List, Tuple, TypeVar, Iterable, Sequence
from functools import reduce, partial
from itertools import chain
from contextlib import contextmanager
from collections import defaultdict

from AnyQt.QtGui import (
    QPen,
    QBrush,
    QColor,
    QPixmap,
    QPainter,
    QTransform,
    QKeySequence,
    QPainterPath,
)
from AnyQt.QtCore import Qt, Slot, QSize, QRectF, QItemSelectionModel
from AnyQt.QtWidgets import (
    QMenu,
    QAction,
    QSplitter,
    QTreeWidget,
    QGraphicsItem,
    QGraphicsView,
    QGraphicsScene,
    QTreeWidgetItem,
    QGraphicsPathItem,
    QGraphicsPixmapItem,
)

import Orange
from Orange.widgets import gui, widget, settings
from Orange.widgets.utils import concurrent

from orangecontrib.bioinformatics import kegg, geneset
from orangecontrib.bioinformatics.utils import statistics
from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID,
    GENE_ID_COLUMN,
    GENE_ID_ATTRIBUTE,
    GENE_AS_ATTRIBUTE_NAME,
    ERROR_ON_MISSING_TAX_ID,
    ERROR_ON_MISSING_GENE_ID,
    ERROR_ON_MISSING_ANNOTATION,
)

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
D = TypeVar("D")


def relation_list_to_multimap(rellist: Sequence[Tuple[A, B]]) -> Dict[A, Sequence[B]]:
    """
    Convert a 'relations' list into a multimap

    Parameters
    ----------
    rellist : Iterable[Tuple[Hashable[K], Any]]

    Returns
    -------
    multimap : Dict[K, List[Any]]

    Example
    -------
    >>> relation_list_to_multimap([(1, "a"), (2, "c"), (1, "b")])
    {1: ['a', 'b'], 2: ['c']}
    """
    mmap = defaultdict(list)
    for key, val in rellist:
        mmap[key].append(val)
    return dict(mmap)


def relation_join(
    rela: Sequence[Tuple[A, B]], relb: Sequence[Tuple[C, D]]
) -> Sequence[Tuple[A, D]]:
    res = []
    relb_mmap = relation_list_to_multimap(relb)
    for a, b in rela:
        res.extend((a, d) for d in relb_mmap.get(b, ()))
    return res


def relation_map(
    relation: Sequence[Tuple[A, B]], values: Sequence[A]
) -> Sequence[Tuple[A, Sequence[B]]]:
    rel_mmap = relation_list_to_multimap(relation)
    return [rel_mmap.get(v, []) for v in values]


def flatten(seq: Iterable[Iterable[A]]) -> Sequence[A]:
    return list(chain.from_iterable(seq))


def split_and_strip(string, sep=None):
    return [s.strip() for s in string.split(sep)]


def path_from_graphics(graphics):
    """
    Return a constructed `QPainterPath` for a KEGG pathway graphics element.
    """
    path = QPainterPath()
    x, y, w, h = [int(graphics.get(c, 0)) for c in ["x", "y", "width", "height"]]
    type = graphics.get("type", "rectangle")
    if type == "rectangle":
        path.addRect(QRectF(x - w / 2, y - h / 2, w, h))
    elif type == "roundrectangle":
        path.addRoundedRect(QRectF(x - w / 2, y - h / 2, w, h), 10, 10)
    elif type == "circle":
        path.addEllipse(QRectF(x - w / 2, y - h / 2, w, h))
    else:
        ValueError("Unknown graphics type %r." % type)
    return path


class EntryGraphicsItem(QGraphicsPathItem):
    """
    An Graphics Item with actions for an overlay of a KEGG pathway image.
    """

    def __init__(self, graphics, *args):
        QGraphicsPathItem.__init__(self, *args)
        path = path_from_graphics(graphics)
        self.setPath(path)
        self.setAcceptHoverEvents(True)
        self._actions = []
        self.link = None

    def hoverEnterEvent(self, event):
        self.setBrush(QBrush(QColor(0, 100, 0, 100)))

    def hoverLeaveEvent(self, event):
        self.setBrush(QBrush(Qt.NoBrush))

    def contextMenuEvent(self, event):
        if self._actions:
            self._menu = menu = QMenu()
            for action in self._actions:
                menu.addAction(action)
            menu.popup(event.screenPos())

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSelectedHasChanged:
            self.setPen(QPen(Qt.red if self.isSelected() else Qt.blue, 2))

        return QGraphicsPathItem.itemChange(self, change, value)


class GraphicsPathwayItem(QGraphicsPixmapItem):
    """
    A Graphics Item displaying a KEGG Pathway image with optional
    marked objects.

    """

    def __init__(self, pathway, objects, *args, **kwargs):
        QGraphicsPixmapItem.__init__(self, *args)
        self.setTransformationMode(Qt.SmoothTransformation)
        self.setPathway(pathway)
        self.setMarkedObjects(objects)

    def setPathway(self, pathway):
        """
        Set pathway
        """
        self.pathway = pathway
        if pathway:
            image_filename = pathway.get_image()
            self._pixmap = QPixmap(image_filename)
        else:
            self._pixmap = QPixmap()
        self.setPixmap(self._pixmap)

    def setMarkedObjects(self, objects):
        for entry in self.pathway.entries() if self.pathway else []:
            if entry.type == "group":
                continue
            graphics = entry.graphics
            contained_objects = [obj for obj in objects if obj in entry.name]
            item = EntryGraphicsItem(graphics, self)
            item.setToolTip(self.tooltip(entry, contained_objects))
            item._actions = self.actions(entry, contained_objects)
            item.marked_objects = contained_objects
            if contained_objects:
                item.setPen(QPen(Qt.blue, 2))
                item.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def actions(self, entry, marked_objects=[]):
        actions = []
        type = entry.type
        if marked_objects:
            action = QAction("View genes on kegg website", None)
            org = {s.split(":")[0] for s in marked_objects}.pop()
            genes = [s.split(":")[-1] for s in marked_objects]
            address = "http://www.genome.jp/dbget-bin/www_bget?" + "+".join(
                [org] + genes
            )

            action.triggered.connect(partial(webbrowser.open, address))
            actions.append(action)
        elif hasattr(entry, "link"):
            action = QAction("View %s on KEGG website" % str(type), None)
            action.triggered.connect(partial(webbrowser.open, entry.link))
            actions.append(action)
        return actions

    def tooltip(self, entry, objects):
        names = [obj for obj in objects if obj in entry.name]
        text = entry.name[:16] + " ..." if len(entry.name) > 20 else entry.name
        text = "<p>%s</p>" % text
        if names:
            text += "<br>".join(names)
        return text

    def contextMenuEvent(self, event):
        self._menu = menu = QMenu()
        action = menu.addAction("View this pathway on KEGG website")
        address = "http://www.kegg.jp/kegg-bin/show_pathway?%s%s" % (
            self.pathway.org,
            self.pathway.number,
        )

        action.triggered.connect(partial(webbrowser.open, address))
        menu.popup(event.screenPos())


class PathwayView(QGraphicsView):
    def __init__(self, master, *args):
        QGraphicsView.__init__(self, *args)
        self.master = master

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        self.setRenderHints(QPainter.Antialiasing)
        scene = QGraphicsScene(self)
        self.pixmapGraphicsItem = QGraphicsPixmapItem(None)
        scene.addItem(self.pixmapGraphicsItem)
        self.setScene(scene)

        self.setMouseTracking(True)
        self.viewport().setMouseTracking(True)

        self.setFocusPolicy(Qt.WheelFocus)

    def SetPathway(self, pathway=None, objects=[]):
        self.scene().clear()
        self.pathway = pathway
        self.objects = objects
        self.pathwayItem = GraphicsPathwayItem(pathway, objects, None)

        self.scene().addItem(self.pathwayItem)
        self.scene().setSceneRect(self.pathwayItem.boundingRect())
        self.updateTransform()

    def resizeEvent(self, event):
        self.updateTransform()
        return QGraphicsView.resizeEvent(self, event)

    def updateTransform(self):
        if self.master.autoResize:
            self.fitInView(
                self.scene().sceneRect().adjusted(-1, -1, 1, 1), Qt.KeepAspectRatio
            )
        else:
            self.setTransform(QTransform())


class OWKEGGPathwayBrowser(widget.OWWidget):
    name = "KEGG Pathways"
    description = "Browse KEGG pathways that include an input set of genes."
    icon = "../widgets/icons/OWKEGGPathwayBrowser.svg"
    priority = 70

    inputs = [
        ("Data", Orange.data.Table, "SetData", widget.Default),
        ("Reference", Orange.data.Table, "SetRefData"),
    ]
    outputs = [
        ("Selected Data", Orange.data.Table, widget.Default),
        ("Unselected Data", Orange.data.Table),
    ]

    autoCommit = settings.Setting(False)
    autoResize = settings.Setting(True)
    useReference = settings.Setting(False)
    showOrthology = settings.Setting(True)

    Ready, Initializing, Running = 0, 1, 2

    class Error(widget.OWWidget.Error):
        missing_annotation = widget.Msg(ERROR_ON_MISSING_ANNOTATION)
        missing_gene_id = widget.Msg(ERROR_ON_MISSING_GENE_ID)
        missing_tax_id = widget.Msg(ERROR_ON_MISSING_TAX_ID)

    def __init__(self, parent=None):
        super().__init__(parent)

        self._changedFlag = False
        self.__invalidated = False
        self.__runstate = OWKEGGPathwayBrowser.Initializing
        self.__in_setProgress = False

        self.controlArea.setMaximumWidth(250)
        box = gui.widgetBox(self.controlArea, "Info")
        self.infoLabel = gui.widgetLabel(box, "No data on input\n")

        gui.separator(self.controlArea)

        gui.checkBox(
            self.controlArea,
            self,
            "useReference",
            "From signal",
            box="Reference",
            callback=self.Update,
        )

        gui.separator(self.controlArea)

        gui.checkBox(
            self.controlArea,
            self,
            "showOrthology",
            "Show pathways in full orthology",
            box="Orthology",
            callback=self.UpdateListView,
        )

        gui.checkBox(
            self.controlArea,
            self,
            "autoResize",
            "Resize to fit",
            box="Image",
            callback=self.UpdatePathwayViewTransform,
        )

        box = gui.widgetBox(self.controlArea, "Cache Control")

        gui.button(
            box,
            self,
            "Clear cache",
            callback=self.ClearCache,
            tooltip="Clear all locally cached KEGG data.",
            default=False,
            autoDefault=False,
        )

        gui.separator(self.controlArea)

        gui.auto_commit(self.controlArea, self, "autoCommit", "Commit")

        gui.rubber(self.controlArea)

        spliter = QSplitter(Qt.Vertical, self.mainArea)
        self.pathwayView = PathwayView(self, spliter)
        self.pathwayView.scene().selectionChanged.connect(self._onSelectionChanged)
        self.mainArea.layout().addWidget(spliter)

        self.listView = QTreeWidget(
            allColumnsShowFocus=True,
            selectionMode=QTreeWidget.SingleSelection,
            sortingEnabled=True,
            maximumHeight=200,
        )

        spliter.addWidget(self.listView)

        self.listView.setColumnCount(4)
        self.listView.setHeaderLabels(["Pathway", "P value", "Genes", "Reference"])

        self.listView.itemSelectionChanged.connect(self.UpdatePathwayView)

        select = QAction("Select All", self, shortcut=QKeySequence.SelectAll)
        select.triggered.connect(self.selectAll)
        self.addAction(select)

        self.data = None
        self.input_genes = []
        self.tax_id = None
        self.use_attr_names = None
        self.gene_id_attribute = None
        self.gene_id_column = None

        self.ref_data = None
        self.ref_genes = []
        self.ref_tax_id = None
        self.ref_use_attr_names = None
        self.ref_gene_id_attribute = None
        self.ref_gene_id_column = None

        self.pathways = {}
        self.ncbi_gene_map = []
        self.org = None

        self._executor = concurrent.ThreadExecutor()
        self.setEnabled(False)
        self.setBlocking(True)
        progress = concurrent.methodinvoke(self, "setProgress", (float,))

        def get_genome():
            """Return a KEGGGenome with the common org entries precached."""
            genome = kegg.KEGGGenome()

            essential = genome.essential_organisms()
            common = genome.common_organisms()
            # Remove duplicates of essential from common.
            # (essential + common list as defined here will be used in the
            # GUI.)
            common = [c for c in common if c not in essential]

            # TODO: Add option to specify additional organisms not
            # in the common list.

            keys = list(map(genome.org_code_to_entry_key, essential + common))

            genome.pre_cache(keys, progress_callback=progress)
            return (keys, genome)

        self._genomeTask = task = concurrent.Task(function=get_genome)
        task.finished.connect(self.__initialize_finish)

        self.progressBarInit()
        self.infoLabel.setText("Fetching organism definitions\n")
        self._executor.submit(task)

    def __initialize_finish(self):
        if self.__runstate != OWKEGGPathwayBrowser.Initializing:
            return

        try:
            keys, genome = self._genomeTask.result()
        except Exception as err:
            self.error(0, str(err))
            raise

        self.progressBarFinished()
        self.setEnabled(True)
        self.setBlocking(False)

        self.infoLabel.setText("No data on input\n")

    def clear(self):
        """
        Clear the widget state.
        """
        self.pathways = {}
        self.org = None

        self.infoLabel.setText("No data on input\n")
        self.listView.clear()
        self.pathwayView.SetPathway(None)

        self.send("Selected Data", None)
        self.send("Unselected Data", None)

    def SetData(self, data=None):
        if self.__runstate == OWKEGGPathwayBrowser.Initializing:
            self.__initialize_finish()

        self.Error.clear()
        if data:
            self.data = data
            self.tax_id = str(self.data.attributes.get(TAX_ID, None))
            self.use_attr_names = self.data.attributes.get(GENE_AS_ATTRIBUTE_NAME, None)
            self.gene_id_attribute = self.data.attributes.get(GENE_ID_ATTRIBUTE, None)
            self.gene_id_column = self.data.attributes.get(GENE_ID_COLUMN, None)

            if not (
                self.use_attr_names is not None
                and ((self.gene_id_attribute is None) ^ (self.gene_id_column is None))
            ):
                if self.tax_id is None:
                    self.Error.missing_annotation()
                    return

                self.Error.missing_gene_id()
                return

            elif self.tax_id is None:
                self.Error.missing_tax_id()
                return

            self.warning(0)
            self.error(0)
            self.information(0)

            self.__invalidated = True
        else:
            self.clear()

    def SetRefData(self, data=None):
        self.information(1)

        if data is not None and self.useReference:
            self.ref_data = data
            self.ref_tax_id = str(self.ref_data.attributes.get(TAX_ID, None))
            self.ref_use_attr_names = self.ref_data.attributes.get(
                GENE_AS_ATTRIBUTE_NAME, None
            )
            self.ref_gene_id_attribute = self.ref_data.attributes.get(
                GENE_ID_ATTRIBUTE, None
            )
            self.ref_gene_id_column = self.ref_data.attributes.get(GENE_ID_COLUMN, None)

            if not (
                self.ref_use_attr_names is not None
                and (
                    (self.ref_gene_id_attribute is None)
                    ^ (self.ref_gene_id_column is None)
                )
            ):
                if self.ref_tax_id is None:
                    self.Error.missing_annotation()
                    return

                self.Error.missing_gene_id()
                return

            elif self.ref_tax_id is None:
                self.Error.missing_tax_id()
                return

            self.__invalidated = True

    def handleNewSignals(self):
        if self.__invalidated:
            self.Update()
            self.__invalidated = False

    def UpdateListView(self):
        self.bestPValueItem = None
        self.listView.clear()
        if not self.data:
            return

        allPathways = self.org.pathways()
        allRefPathways = kegg.pathways("map")

        items = []
        kegg_pathways = kegg.KEGGPathways()

        org_code = self.org.org_code

        if self.showOrthology:
            self.koOrthology = kegg.KEGGBrite("ko00001")
            self.listView.setRootIsDecorated(True)
            path_ids = {s[-5:] for s in self.pathways.keys()}

            def _walkCollect(koEntry):
                num = koEntry.title[:5] if koEntry.title else None
                if num in path_ids:
                    return [koEntry] + reduce(
                        lambda li, c: li + _walkCollect(c),
                        [child for child in koEntry.entries],
                        [],
                    )
                else:
                    c = reduce(
                        lambda li, c: li + _walkCollect(c),
                        [child for child in koEntry.entries],
                        [],
                    )
                    return c + (c and [koEntry] or [])

            allClasses = reduce(
                lambda li1, li2: li1 + li2,
                [_walkCollect(c) for c in self.koOrthology],
                [],
            )

            def _walkCreate(koEntry, lvItem):
                item = QTreeWidgetItem(lvItem)
                id = "path:" + org_code + koEntry.title[:5]

                if koEntry.title[:5] in path_ids:
                    p = kegg_pathways.get_entry(id)
                    if p is None:
                        # In case the genesets still have obsolete entries
                        name = koEntry.title
                    else:
                        name = p.name
                    genes, p_value, ref = self.pathways[id]
                    item.setText(0, name)
                    item.setText(1, "%.5f" % p_value)
                    item.setText(2, "%i of %i" % (len(genes), len(self.input_genes)))
                    item.setText(3, "%i of %i" % (ref, len(self.ref_genes)))
                    item.pathway_id = id if p is not None else None
                else:
                    if id in allPathways:
                        text = kegg_pathways.get_entry(id).name
                    else:
                        text = koEntry.title
                    item.setText(0, text)

                    if id in allPathways:
                        item.pathway_id = id
                    elif "path:map" + koEntry.title[:5] in allRefPathways:
                        item.pathway_id = "path:map" + koEntry.title[:5]
                    else:
                        item.pathway_id = None

                for child in koEntry.entries:
                    if child in allClasses:
                        _walkCreate(child, item)

            for koEntry in self.koOrthology:
                if koEntry in allClasses:
                    _walkCreate(koEntry, self.listView)

            self.listView.update()
        else:
            self.listView.setRootIsDecorated(False)
            pathways = self.pathways.items()
            pathways = sorted(pathways, key=lambda item: item[1][1])

            for id, (genes, p_value, ref) in pathways:
                item = QTreeWidgetItem(self.listView)
                item.setText(0, kegg_pathways.get_entry(id).name)
                item.setText(1, "%.5f" % p_value)
                item.setText(2, "%i of %i" % (len(genes), len(self.input_genes)))
                item.setText(3, "%i of %i" % (ref, len(self.ref_genes)))
                item.pathway_id = id
                items.append(item)

        self.bestPValueItem = items and items[0] or None
        self.listView.expandAll()
        for i in range(4):
            self.listView.resizeColumnToContents(i)

        if self.bestPValueItem:
            index = self.listView.indexFromItem(self.bestPValueItem)
            self.listView.selectionModel().select(
                index, QItemSelectionModel.ClearAndSelect
            )

    def UpdatePathwayView(self):
        items = self.listView.selectedItems()

        if len(items) > 0:
            item = items[0]
        else:
            item = None

        self.commit()
        item = item or self.bestPValueItem
        if not item or not item.pathway_id:
            self.pathwayView.SetPathway(None)
            return

        def get_kgml_and_image(pathway_id):
            """Return an initialized KEGGPathway with pre-cached data"""
            p = kegg.KEGGPathway(pathway_id)
            p._get_kgml()  # makes sure the kgml file is downloaded
            p._get_image_filename()  # makes sure the image is downloaded
            return (pathway_id, p)

        self.setEnabled(False)
        self._pathwayTask = concurrent.Task(
            function=lambda: get_kgml_and_image(item.pathway_id)
        )
        self._pathwayTask.finished.connect(self._onPathwayTaskFinshed)
        self._executor.submit(self._pathwayTask)

    def _onPathwayTaskFinshed(self):
        self.setEnabled(True)
        pathway_id, self.pathway = self._pathwayTask.result()
        objects = self.pathways.get(pathway_id, [[]])[0]  # [ncbi_gene_id]
        # map ncbi_gene_id to keg_id for display
        objects = flatten(relation_map(self.ncbi_gene_map, objects))
        self.pathwayView.SetPathway(self.pathway, objects)

    def UpdatePathwayViewTransform(self):
        self.pathwayView.updateTransform()

    def Update(self):
        """
        Update (recompute enriched pathways) the widget state.
        """
        if not self.data:
            return

        self.error(0)
        self.information(0)

        # XXX: Check data in setData, do not even allow this to be executed if
        # data has no genes
        try:
            self.__get_input_genes()
            self.input_genes = set(self.input_genes)
        except ValueError:
            self.error(0, "Cannot extract gene names from input.")

        self.information(1)

        self.org = kegg.KEGGOrganism(kegg.from_taxid(self.tax_id))

        if self.useReference and self.ref_data:
            self.__get_ref_genes()
            self.ref_genes = set(self.ref_genes)
        else:
            self.ref_genes = self.org.get_ncbi_ids()

        def run_enrichment(org_code, genes, reference, progress=None):
            # We use the kegg pathway gene sets provided by 'geneset' for
            # the enrichment calculation.

            kegg_api = kegg.api.CachedKeggApi()
            link_map = kegg_api.link(
                org_code, "pathway"
            )  # [(pathway_id, kegg_gene_id)]
            ncbi_gene_map = kegg_api.conv(
                org_code, 'ncbi-geneid'
            )  # [(ncbi_gene_id, kegg_gene_id)]
            ncbi_gene_map = [(_1.split(":", 1)[1], _2) for _1, _2 in ncbi_gene_map]
            link_map = relation_join(
                link_map, [(_2, _1) for _1, _2 in ncbi_gene_map]
            )  # [(pathway_id, ncbi_gene_id)]
            kegg_sets = relation_list_to_multimap(
                link_map
            )  #  {pathway_id -> [ncbi_gene_ids]}
            # map kegg gene ids to ncbi_gene_ids.
            kegg_sets = geneset.GeneSets(
                sets=[
                    geneset.GeneSet(gs_id=ddi, genes=set(genes))
                    for ddi, genes in kegg_sets.items()
                ]
            )
            pathways = pathway_enrichment(
                kegg_sets, genes, reference, callback=progress
            )
            # Ensure that pathway entries are pre-cached for later use in the
            # list/tree view
            kegg_pathways = kegg.KEGGPathways()
            kegg_pathways.pre_cache(pathways.keys(), progress_callback=progress)

            return pathways, ncbi_gene_map

        self.progressBarInit()
        self.setEnabled(False)
        self.infoLabel.setText("Retrieving...\n")

        progress = concurrent.methodinvoke(self, "setProgress", (float,))

        run_func = partial(
            run_enrichment,
            self.org.org_code,
            self.input_genes,
            self.ref_genes,
            progress,
        )
        self._enrichTask = concurrent.Task(function=run_func)
        self._enrichTask.finished.connect(self._onEnrichTaskFinished)
        self._executor.submit(self._enrichTask)

    def _onEnrichTaskFinished(self):
        self.setEnabled(True)
        self.setBlocking(False)
        try:
            pathways, ncbi_gene_map = self._enrichTask.result()
        except Exception:
            raise

        self.progressBarFinished()

        self.pathways = pathways
        self.ncbi_gene_map = ncbi_gene_map

        if not self.pathways:
            self.warning(0, "No enriched pathways found.")
        else:
            self.warning(0)

        self.infoLabel.setText(
            "{} unique gene names on input\n".format(len(set(self.input_genes)))
        )

        self.UpdateListView()

    @Slot(float)
    def setProgress(self, value):
        if self.__in_setProgress:
            return

        self.__in_setProgress = True
        self.progressBarSet(value)
        self.__in_setProgress = False

    def __get_input_genes(self):
        """
        Extract and return gene names from `data`.
        """
        self.input_genes = []

        if self.use_attr_names:
            for variable in self.data.domain.attributes:
                self.input_genes.append(
                    str(variable.attributes.get(self.gene_id_attribute, '?'))
                )
        else:
            genes = self.data.get_column(self.gene_id_column)
            self.input_genes = [str(g) for g in genes]

        if len(self.input_genes) <= 0:
            raise ValueError("No gene names in data.")

    def __get_ref_genes(self):
        """
        Extract and return gene names from `data`.
        """
        self.ref_genes = []

        if self.ref_use_attr_names:
            for variable in self.ref_data.domain.attributes:
                self.ref_genes.append(
                    str(variable.attributes.get(self.ref_gene_id_attribute, '?'))
                )
        else:
            genes = self.ref_data.get_column(self.ref_gene_id_column)
            self.ref_genes = [str(g) for g in genes]

    def selectAll(self):
        """
        Select all items in the pathway view.
        """
        changed = False
        scene = self.pathwayView.scene()
        with disconnected(scene.selectionChanged, self._onSelectionChanged):
            for item in scene.items():
                if (
                    item.flags() & QGraphicsItem.ItemIsSelectable
                    and not item.isSelected()
                ):
                    item.setSelected(True)
                    changed = True
        if changed:
            self._onSelectionChanged()

    def _onSelectionChanged(self):
        # Item selection in the pathwayView/scene has changed
        self.commit()

    def commit(self):
        if self.data:
            selectedItems = self.pathwayView.scene().selectedItems()
            selectedGenes = reduce(
                set.union, [item.marked_objects for item in selectedItems], set()
            )
            # map kegg_ids back to ncbi_gene_id
            backmap = [(_2, _1) for _1, _2 in self.ncbi_gene_map]
            selectedGenes = set(flatten(relation_map(backmap, selectedGenes)))
            if self.use_attr_names:
                selected = [
                    column
                    for column in self.data.domain.attributes
                    if self.gene_id_attribute in column.attributes
                    and str(column.attributes[self.gene_id_attribute]) in selectedGenes
                ]
                data = self.data[:, selected]
                self.send("Selected Data", data)
            else:
                selected_indices = []
                other_indices = []

                for row_index, row in enumerate(self.data):
                    gene_in_row = str(row[self.gene_id_column])
                    if gene_in_row in self.input_genes and gene_in_row in selectedGenes:
                        selected_indices.append(row_index)
                    else:
                        other_indices.append(row_index)

                if selected_indices:
                    selected = self.data[selected_indices]
                else:
                    selected = None

                if other_indices:
                    other = self.data[other_indices]
                else:
                    other = None

                self.send("Selected Data", selected)
                self.send("Unselected Data", other)
        else:
            self.send("Selected Data", None)
            self.send("Unselected Data", None)

    def ClearCache(self):
        kegg.caching.clear_cache()

    def onDeleteWidget(self):
        """
        Called before the widget is removed from the canvas.
        """
        super().onDeleteWidget()

        self.org = None
        self._executor.shutdown(wait=False)
        gc.collect()  # Force collection (WHY?)

    def sizeHint(self):
        return QSize(1024, 720)


def pathway_enrichment(genesets, genes, reference, prob=None, callback=None):
    result_sets = []
    p_values = []
    if prob is None:
        prob = statistics.Hypergeometric()

    for i, gs in enumerate(genesets):
        cluster = gs.genes.intersection(genes)
        ref = gs.genes.intersection(reference)
        k = len(cluster)
        N = len(reference)
        m = len(ref)
        n = len(genes)
        if k:
            p_val = prob.p_value(k, N, m, n)
            result_sets.append((gs.gs_id, cluster, ref))
            p_values.append(p_val)
        if callback is not None:
            callback(100.0 * i / len(genesets))

    # FDR correction
    p_values = statistics.FDR(p_values)

    return {
        _id: (genes, p_val, len(ref))
        for (_id, genes, ref), p_val in zip(result_sets, p_values)
    }


@contextmanager
def disconnected(signal, slot):
    """
    A context manager disconnecting a slot from a signal.
    ::

        with disconnected(scene.selectionChanged, self.onSelectionChanged):
            # Can change item selection in a scene without
            # onSelectionChanged being invoked.
            do_something()

    """
    signal.disconnect(slot)
    try:
        yield
    finally:
        signal.connect(slot)


if __name__ == "__main__":

    def test_main(argv=sys.argv):
        from AnyQt.QtWidgets import QApplication

        app = QApplication(sys.argv)

        if len(argv) > 1:
            filename = argv[1]
        else:
            filename = "brown-selected"
        data = Orange.data.Table(filename)

        w = OWKEGGPathwayBrowser()
        w.show()
        w.raise_()

        w.SetData(data)
        w.handleNewSignals()

        r = app.exec_()
        w.saveSettings()
        w.onDeleteWidget()
        return r

    sys.exit(test_main())



import unittest
from unittest.mock import patch

from AnyQt.QtTest import QSignalSpy

from Orange.widgets.tests.base import GuiTest, WidgetTest
from Orange.widgets.credentials import CredentialManager
from Orange.widgets.tests.utils import simulate

from orangecontrib.bioinformatics.resolwe import ResolweAuthError
from orangecontrib.bioinformatics.widgets.components.resolwe import SignIn
from orangecontrib.bioinformatics.widgets.OWGenialisExpressions import SortBy, ItemsPerPage, FilterByDateModified
from orangecontrib.bioinformatics.tests.widgets.test_OWGenialisExpressions import MockWidget


class TestSignInForm(GuiTest):
    @patch(
        'orangecontrib.bioinformatics.widgets.components.resolwe.get_credential_manager',
        return_value=CredentialManager('resolwe_credentials_test'),
    )
    @patch('orangecontrib.bioinformatics.widgets.components.resolwe.connect')
    def test_dialog_success(self, mocked_connect, mocked_cm):
        widget = MockWidget()
        dialog = SignIn(widget)
        self.assertFalse(dialog.sign_in_btn.isEnabled())
        dialog.sign_in()
        mocked_connect.assert_not_called()
        self.assertIsNone(dialog.resolwe_instance)

        dialog.username_line_edit.setText('foo')
        dialog.sign_in()
        mocked_connect.assert_not_called()

        dialog.password_line_edit.setText('bar')
        self.assertTrue(dialog.sign_in_btn.isEnabled())
        dialog.sign_in()

        mocked_connect.assert_called_once()
        self.assertTrue(dialog.error_msg.isHidden())
        self.assertIsNotNone(dialog.resolwe_instance)

        # cleanup
        cm = CredentialManager('resolwe_credentials_test')
        if cm.username:
            del cm.username
        if cm.password:
            del cm.password

    @patch(
        'orangecontrib.bioinformatics.widgets.components.resolwe.get_credential_manager',
        return_value=CredentialManager('resolwe_credentials_test'),
    )
    @patch('orangecontrib.bioinformatics.widgets.components.resolwe.connect', side_effect=ResolweAuthError())
    def test_dialog_fail(self, mocked_connect, mocked_cm):
        widget = MockWidget()
        dialog = SignIn(widget)
        dialog.username_line_edit.setText('foo')
        dialog.password_line_edit.setText('bar')

        self.assertTrue(dialog.error_msg.isHidden())
        dialog.sign_in()
        self.assertFalse(dialog.error_msg.isHidden())

        mocked_connect.assert_called_once()


class TestCollapsibleFilterComponent(WidgetTest):
    def setUp(self):
        self.widget = MockWidget()
        self.component = self.widget.filter_component
        self.emitted_signals: QSignalSpy = QSignalSpy(self.component.options_changed)

    def test_filter_by_modified(self):
        # check if correct option is set by default
        self.assertEqual(FilterByDateModified.any_time, self.component.filter_by_modified)
        # run trough all the options
        simulate.combobox_run_through_all(self.component.filter_modified)
        self.assertEqual(FilterByDateModified.past_month, self.component.filter_by_modified)

        # check if signal is emitted on every option changed
        self.assertEqual(len(FilterByDateModified.labels()), len(self.emitted_signals))

    def test_sorting(self):
        # check if correct option is set by default
        self.assertEqual(SortBy.newest_first, self.component.sort_by)
        # run trough all the options
        simulate.combobox_run_through_all(self.component.sorting)
        self.assertEqual(SortBy.oldest_first, self.component.sort_by)

        # check if signal is emitted on every option changed
        self.assertEqual(len(SortBy.labels()), len(self.emitted_signals))

    def test_full_text_search(self):
        search_parameter = 'foobar'
        # Default sorting should be set initially
        self.assertEqual(self.component.sort_by, SortBy.newest_first)
        # simulate user input
        self.component.filter_full_text.setText(search_parameter)
        self.component.on_filter_full_text_changed()
        # check if setting is correctly set
        self.assertEqual(self.component.filter_by_full_text, search_parameter)
        # check if sorting is set accordingly
        self.assertEqual(SortBy.relevance, self.component.sort_by)


class TestPaginationComponent(WidgetTest):
    def setUp(self):
        self.widget = MockWidget()
        self.component = self.widget.pagination_component

    def test_pagination_availability(self):
        self.widget.pagination_availability.emit(False, False)
        self.assertFalse(self.component.page_left_btn.isEnabled())
        self.assertFalse(self.component.page_right_btn.isEnabled())

        self.widget.pagination_availability.emit(True, False)
        self.assertFalse(self.component.page_left_btn.isEnabled())
        self.assertTrue(self.component.page_right_btn.isEnabled())

        self.widget.pagination_availability.emit(False, True)
        self.assertTrue(self.component.page_left_btn.isEnabled())
        self.assertFalse(self.component.page_right_btn.isEnabled())

        self.widget.pagination_availability.emit(True, True)
        self.assertTrue(self.component.page_left_btn.isEnabled())
        self.assertTrue(self.component.page_right_btn.isEnabled())

    def test_page_limit(self):
        # test default option
        self.assertTrue(self.component.page_limit.buttons[ItemsPerPage.min].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.med].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.max].isChecked())
        self.assertEqual(0, self.component.offset)
        self.assertEqual(1, self.component.current_page)

        # change current page
        self.component.right_btn_pressed()
        self.assertEqual(ItemsPerPage.values()[ItemsPerPage.min], self.component.offset)
        self.assertEqual(2, self.component.current_page)

        # change page limit
        self.component.items_per_page = ItemsPerPage.max
        self.component.on_limit_changed()
        self.assertEqual(0, self.component.offset)
        self.assertEqual(1, self.component.current_page)
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.min].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.med].isChecked())
        self.assertTrue(self.component.page_limit.buttons[ItemsPerPage.max].isChecked())

        # change current page
        self.component.right_btn_pressed()
        self.assertEqual(ItemsPerPage.values()[ItemsPerPage.max], self.component.offset)
        self.assertEqual(2, self.component.current_page)

    def test_pagination(self):
        # change page limit
        self.component.items_per_page = ItemsPerPage.med
        self.component.on_limit_changed()
        self.assertEqual(0, self.component.offset)
        self.assertEqual(1, self.component.current_page)
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.min].isChecked())
        self.assertTrue(self.component.page_limit.buttons[ItemsPerPage.med].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.max].isChecked())

        current_offset = 0

        for page in [2, 3, 4]:
            self.component.right_btn_pressed()
            current_offset += ItemsPerPage.values()[ItemsPerPage.med]
            self.assertEqual(current_offset, self.component.offset)
            self.assertEqual(page, self.component.current_page)

        for page in [3, 2, 1]:
            self.component.left_btn_pressed()
            current_offset -= ItemsPerPage.values()[ItemsPerPage.med]
            self.assertEqual(current_offset, self.component.offset)
            self.assertEqual(page, self.component.current_page)

        # test left pagination limit
        for _ in range(3):
            self.component.left_btn_pressed()
            self.assertEqual(0, self.component.offset)
            self.assertEqual(1, self.component.current_page)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from AnyQt.QtCore import QModelIndex, QItemSelectionModel

from Orange.data import Table
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import simulate

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.OWMarkerGenes import (
    SERVER_FILES_DOMAIN,
    TreeItem,
    TreeView,
    OWMarkerGenes,
)


class TestTreeItem(unittest.TestCase):
    def setUp(self) -> None:
        self.item1 = TreeItem("item 1", False, None, None)
        self.item2 = TreeItem("item 2", False, None, None)
        self.item3 = TreeItem("item 3", False, None, None)

    def test_change_parent(self):
        self.item1.change_parent(self.item2)
        self.assertEqual(self.item1.parentItem, self.item2)
        self.assertIn(self.item1, self.item2.childItems)

    def test_change_parent_none(self):
        self.item1.change_parent(None)
        self.assertEqual(self.item1.parentItem, None)
        self.assertListEqual([], self.item2.childItems)

    def change_parents(self):
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertEqual(self.item1.parentItem, self.item3)
        self.assertIn(self.item1, self.item3.childItems)
        self.assertEqual(self.item2.parentItem, self.item3)
        self.assertIn(self.item2, self.item3.childItems)

    def test_append_child(self):
        self.item1.append_child(self.item3)
        self.assertIn(self.item3, self.item1.childItems)

    def test_remove_children(self):
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(
            -1, 1
        )  # should return False (it is not possible)
        self.assertFalse(s)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(2, 1)  # should return False (it is not possible)
        self.assertFalse(s)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(1, 2)  # should return False (it is not possible)
        self.assertFalse(s)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(0, 1)
        self.assertTrue(s)
        self.assertListEqual([self.item2], self.item3.childItems)

        s = self.item3.remove_children(0, 1)
        self.assertTrue(s)
        self.assertListEqual([], self.item3.childItems)

        s = self.item3.remove_children(0, 1)
        self.assertFalse(s)
        self.assertListEqual([], self.item3.childItems)

        # test remove multiple
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        s = self.item3.remove_children(0, 2)
        self.assertTrue(s)
        self.assertListEqual([], self.item3.childItems)

    def test_children_count(self):
        self.assertEqual(0, self.item3.children_count())

        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertEqual(2, self.item3.children_count())

        self.item3.remove_children(0, 1)
        self.assertEqual(1, self.item3.children_count())

        self.item3.remove_children(0, 1)
        self.assertEqual(0, self.item3.children_count())

    def test_row(self):
        # when no parent it is assumed that the item is a root item - its row is 0
        self.assertEqual(0, self.item1.row())
        self.assertEqual(0, self.item2.row())
        self.assertEqual(0, self.item3.row())

        # add parent
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertEqual(0, self.item1.row())
        self.assertEqual(1, self.item2.row())
        self.assertEqual(0, self.item3.row())

        # different order
        self.item3.remove_children(0, 2)
        TreeItem.change_parents([self.item2, self.item1], self.item3)
        self.assertEqual(1, self.item1.row())
        self.assertEqual(0, self.item2.row())
        self.assertEqual(0, self.item3.row())

    def test_child_from_name_index(self):
        self.assertTupleEqual((None, None), self.item3.child_from_name_index("item 1"))
        TreeItem.change_parents([self.item1, self.item2], self.item3)

        self.assertTupleEqual(
            (0, self.item1), self.item3.child_from_name_index("item 1")
        )
        self.assertTupleEqual(
            (1, self.item2), self.item3.child_from_name_index("item 2")
        )

        self.item3.remove_children(0, 2)
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.assertTupleEqual(
            (1, self.item1), self.item3.child_from_name_index("item 1")
        )
        self.assertTupleEqual(
            (0, self.item2), self.item3.child_from_name_index("item 2")
        )

    def test_contains_text(self):
        # simulate RowItem with dictionary
        self.item1.data_row = {
            "Name": "item1",
            "Cell Type": "type1",
            "Function": "function1",
        }
        self.item2.data_row = {
            "Name": "item2",
            "Cell Type": "type2",
            "Function": "function2",
        }
        filter_columns = ["Name", "Cell Type", "Function"]
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.assertTrue(self.item1.contains_text("item1", filter_columns))
        self.assertFalse(self.item2.contains_text("item1", filter_columns))
        self.assertFalse(self.item3.contains_text("item1", filter_columns))

        self.assertTrue(self.item1.contains_text("type1", filter_columns))
        self.assertFalse(self.item2.contains_text("type1", filter_columns))
        self.assertFalse(self.item3.contains_text("type1", filter_columns))

        self.assertTrue(self.item1.contains_text("function1", filter_columns))
        self.assertFalse(self.item2.contains_text("function1", filter_columns))
        self.assertFalse(self.item3.contains_text("function1", filter_columns))

        self.assertFalse(self.item1.contains_text("item2", filter_columns))
        self.assertTrue(self.item2.contains_text("item2", filter_columns))
        self.assertFalse(self.item3.contains_text("item2", filter_columns))

        self.assertFalse(self.item1.contains_text("type2", filter_columns))
        self.assertTrue(self.item2.contains_text("type2", filter_columns))
        self.assertFalse(self.item3.contains_text("type2", filter_columns))

        self.assertFalse(self.item1.contains_text("function2", filter_columns))
        self.assertTrue(self.item2.contains_text("function2", filter_columns))
        self.assertFalse(self.item3.contains_text("function2", filter_columns))

        # test case sensitivity
        self.assertTrue(self.item1.contains_text("Item1", filter_columns))
        self.assertFalse(self.item2.contains_text("Item1", filter_columns))

        self.assertTrue(self.item1.contains_text("Type1", filter_columns))
        self.assertFalse(self.item2.contains_text("Type1", filter_columns))

        self.assertFalse(self.item1.contains_text("Item2", filter_columns))
        self.assertTrue(self.item2.contains_text("Item2", filter_columns))

        self.assertFalse(self.item1.contains_text("Type2", filter_columns))
        self.assertTrue(self.item2.contains_text("Type2", filter_columns))

    def test_get_data_rows(self):
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = "dr3"

        self.assertListEqual(
            sorted(["dr1", "dr2", "dr3"]), sorted(self.item3.get_data_rows())
        )
        self.assertListEqual(["dr1"], self.item1.get_data_rows())
        self.assertListEqual(["dr2"], self.item2.get_data_rows())

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = None

        self.assertListEqual(sorted(["dr1", "dr2"]), sorted(self.item3.get_data_rows()))
        self.assertListEqual(["dr1"], self.item1.get_data_rows())
        self.assertListEqual(["dr2"], self.item2.get_data_rows())

        self.item1.data_row = "dr1"
        self.item2.data_row = None
        self.item3.data_row = None

        self.assertListEqual(["dr1"], self.item3.get_data_rows())
        self.assertListEqual(["dr1"], self.item1.get_data_rows())
        self.assertListEqual([], self.item2.get_data_rows())

    def test_count_marker_genes(self):
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = "dr3"

        self.assertEqual(3, self.item3.count_marker_genes())
        self.assertEqual(1, self.item1.count_marker_genes())
        self.assertEqual(1, self.item2.count_marker_genes())

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = None

        self.assertEqual(2, self.item3.count_marker_genes())
        self.assertEqual(1, self.item1.count_marker_genes())
        self.assertEqual(1, self.item2.count_marker_genes())

        self.item1.data_row = "dr1"
        self.item2.data_row = None
        self.item3.data_row = None

        self.assertEqual(1, self.item3.count_marker_genes())
        self.assertEqual(1, self.item1.count_marker_genes())
        self.assertEqual(0, self.item2.count_marker_genes())


class TestOWMarkerGenes(WidgetTest):
    @classmethod
    def setUpClass(cls):
        """Code executed only once for all tests"""
        super().setUpClass()
        file_name = "panglao_gene_markers.tab"
        serverfiles.update(SERVER_FILES_DOMAIN, file_name)
        file_path = serverfiles.localpath_download(SERVER_FILES_DOMAIN, file_name)
        cls.panglao = Table.from_file(file_path)

        file_name = "cellMarker_gene_markers.tab"
        serverfiles.update(SERVER_FILES_DOMAIN, file_name)
        file_path = serverfiles.localpath_download(SERVER_FILES_DOMAIN, file_name)
        cls.cell_markers = Table.from_file(file_path)

    def test_minimum_size(self):
        pass

    def setUp(self):
        self.widget = self.create_widget(OWMarkerGenes)

    def move_top_element(self, source_view: TreeView) -> None:
        """
        This function moves top element form the source_view to another.

        Parameters
        ----------
        source_view
            The view from which we move top element.
        """
        parent_index = source_view.model().index(0, 0)
        source_view.selectionModel().select(
            parent_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()

    def move_first_child(self, source_view: TreeView) -> None:
        """
        This function moves the first child element from source_view to another.

        Parameters
        ----------
        source_view
            The view from which we move top element.
        """
        parent_index = source_view.model().index(0, 0)
        gene_index = source_view.model().index(0, 0, parent_index)
        source_view.selectionModel().select(
            gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()

    def test_data_not_empty(self):
        """
        Test that data are present
        """
        w = self.widget
        self.assertIsNotNone(w.data)

    def test_available_sources(self):
        """
        Test if all sources retrieved. Currently there are two sources available.
        When more available unittest will be changed.
        """
        self.assertListEqual(
            ["CellMarker", "Panglao"], list(self.widget.available_sources.keys())
        )

    def test_source_changed(self):
        """
        Test changing the source of the data.
        """
        # cell marker data
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertTrue(isinstance(self.widget.data, Table))
        self.assertEqual(len(self.panglao), len(self.widget.data))

        # Panglao data
        simulate.combobox_activate_index(self.widget.controls.source_index, 1)
        self.assertEqual("CellMarker", self.widget.db_source_cb.currentText())
        self.assertTrue(isinstance(self.widget.data, Table))
        self.assertEqual(len(self.cell_markers), len(self.widget.data))

        # cell marker data
        simulate.combobox_activate_index(self.widget.controls.source_index, 0)
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertTrue(isinstance(self.widget.data, Table))
        self.assertEqual(len(self.panglao), len(self.widget.data))

    def test_organism_changed(self):
        """
        Test changing organisms in combination with changing the source.
        """
        simulate.combobox_activate_index(self.widget.controls.source_index, 0)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 0)
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertEqual("Human", self.widget.group_cb.currentText())

        human_rows = self.widget.data.get_column("Organism") == "Human"
        cell_types = self.widget.data.get_column("Cell Type")

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(
            sum(self.panglao.get_column("Organism") == "Human"), len(model)
        )
        self.assertEqual(
            len(np.unique(cell_types[human_rows])), len(model.rootItem.childItems)
        )

        simulate.combobox_activate_index(self.widget.controls.source_index, 0)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 1)
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertEqual("Mouse", self.widget.group_cb.currentText())

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(
            sum(self.panglao.get_column("Organism") == "Mouse"), len(model)
        )
        self.assertEqual(
            len(np.unique(cell_types[~human_rows])), len(model.rootItem.childItems)
        )

        simulate.combobox_activate_index(self.widget.controls.source_index, 1)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 0)
        self.assertEqual("CellMarker", self.widget.db_source_cb.currentText())
        self.assertEqual("Human", self.widget.group_cb.currentText())

        human_rows = self.widget.data.get_column("Organism") == "Human"
        cell_types = self.widget.data.get_column("Cell Type")

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(
            sum(self.cell_markers.get_column("Organism") == "Human"), len(model)
        )
        self.assertEqual(
            len(np.unique(cell_types[human_rows])), len(model.rootItem.childItems)
        )

        simulate.combobox_activate_index(self.widget.controls.source_index, 1)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 1)
        self.assertEqual("CellMarker", self.widget.db_source_cb.currentText())
        self.assertEqual("Mouse", self.widget.group_cb.currentText())

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(
            sum(self.cell_markers.get_column("Organism") == "Mouse"), len(model)
        )
        self.assertEqual(
            len(np.unique(cell_types[~human_rows])), len(model.rootItem.childItems)
        )

    def test_group_by_changed(self):
        """
        Test changing the group by parameter (Cell Type, Function)
        """
        # group by cell type
        simulate.combobox_activate_index(
            self.widget.controls.selected_root_attribute, 0
        )

        human_rows = self.widget.data.get_column("Organism") == "Human"
        cell_types = np.unique(self.widget.data.get_column("Cell Type")[human_rows])
        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(len(cell_types), len(model.rootItem.childItems))

        # group by function
        simulate.combobox_activate_index(
            self.widget.controls.selected_root_attribute, 1
        )

        functions = np.unique(self.widget.data.get_column("Function")[human_rows])
        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(len(functions), len(model.rootItem.childItems))

    def test_move_elements(self):
        """
        Move elements between booth views and observe if numbers matches.
        This test just move the single element.
        """
        # select first three elements
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view
        num_items_source_view = len(src_view.model().sourceModel())
        num_root_items = src_view.model().rowCount(QModelIndex())

        # select first element
        index = src_view.model().index(0, 0)
        src_view.selectionModel().select(
            index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        num_items = src_view.model().rowCount(
            index
        )  # number of items under selected item

        # move element to the TreeView with selected genes
        self.widget.move_button.click()
        self.assertEqual(len(dest_view.model().sourceModel()), num_items)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(num_items, len(output))
        self.assertEqual(
            dest_view.model().rowCount(QModelIndex()), 1
        )  # one root element added
        index = dest_view.model().index(0, 0)
        self.assertEqual(dest_view.model().rowCount(index), num_items)

        self.assertEqual(
            src_view.model().rowCount(QModelIndex()), num_root_items - 1
        )  # one item is remove from view

        # move everything back
        dest_view.selectionModel().select(
            index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()

        self.assertEqual(num_items_source_view, len(src_view.model().sourceModel()))
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

    def test_move_more_elements(self):
        """
        Move elements between booth views and observe if numbers matches.
        This test keep moving more elements.
        """
        # select first three elements
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view
        num_items_source_view = len(src_view.model().sourceModel())
        num_root_items = src_view.model().rowCount(QModelIndex())

        # select first element
        index = src_view.model().index(0, 0)
        index1 = src_view.model().index(1, 0)
        src_view.selectionModel().select(
            index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        src_view.selectionModel().select(
            index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        num_items = src_view.model().rowCount(index)
        num_items1 = src_view.model().rowCount(index1)

        # move elements to the TreeView with selected genes
        self.widget.move_button.click()
        self.assertEqual(len(dest_view.model().sourceModel()), num_items + num_items1)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(num_items + num_items1, len(output))
        self.assertEqual(
            dest_view.model().rowCount(QModelIndex()), 2
        )  # two root element added
        index = dest_view.model().index(0, 0)
        index1 = dest_view.model().index(1, 0)
        self.assertEqual(
            dest_view.model().rowCount(index) + dest_view.model().rowCount(index1),
            num_items + num_items1,
        )

        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items - 2)

        # move everything back one by one
        dest_view.selectionModel().select(
            index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()
        index1 = dest_view.model().index(0, 0)
        self.assertEqual(dest_view.model().rowCount(index1), num_items1)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(num_items1, len(output))
        self.assertEqual(len(dest_view.model().sourceModel()), num_items1)
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items - 1)

        # move the other one back
        dest_view.selectionModel().select(
            index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()
        self.assertEqual(num_items_source_view, len(src_view.model().sourceModel()))
        self.assertEqual(0, len(dest_view.model().sourceModel()))
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)
        self.assertEqual(dest_view.model().rowCount(QModelIndex()), 0)

    def test_move_genes(self):
        """
        Move elements between booth views and observe if numbers matches.
        In previous test we were moving just root elements.
        In this test we will move children elements.
        """
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view
        num_items_source_view = len(src_view.model().sourceModel())
        num_root_items = src_view.model().rowCount(QModelIndex())

        # select first element
        parent_index = src_view.model().index(0, 0)
        index = src_view.model().index(0, 0, parent_index)
        index1 = src_view.model().index(1, 0, parent_index)

        src_view.selectionModel().select(
            index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        src_view.selectionModel().select(
            index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )

        # move elements to the TreeView with selected genes
        self.widget.move_button.click()
        self.assertEqual(len(dest_view.model().sourceModel()), 2)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(2, len(output))
        self.assertEqual(
            dest_view.model().rowCount(QModelIndex()), 1
        )  # one parent element added
        parent_index = dest_view.model().index(0, 0)
        self.assertEqual(dest_view.model().rowCount(parent_index), 2)
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

        # move everything back one by one
        index = dest_view.model().index(0, 0, parent_index)
        dest_view.selectionModel().select(
            index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()

        self.assertEqual(dest_view.model().rowCount(parent_index), 1)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(1, len(output))
        self.assertEqual(len(dest_view.model().sourceModel()), 1)
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

        # move the other one back
        index1 = dest_view.model().index(0, 0, parent_index)
        dest_view.selectionModel().select(
            index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()
        self.assertEqual(num_items_source_view, len(src_view.model().sourceModel()))
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

    def test_description(self):
        """
        Test correctness of the gene description showing in the bottom of widget.
        """
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view

        not_gene = "Select a gene to see information."
        first_gene_info = "Gene name: AKR1C3"

        # first test in the source view
        parent_index = src_view.model().index(0, 0)
        gene_index = src_view.model().index(0, 0, parent_index)
        gene_index1 = src_view.model().index(1, 0, parent_index)

        # no description for parent
        src_view.selectionModel().select(
            parent_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # no description for multiple genes
        src_view.selectionModel().select(
            gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        src_view.selectionModel().select(
            gene_index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # description for first gene
        src_view.selectionModel().select(
            gene_index, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
        )
        self.assertIn(first_gene_info, self.widget.descriptionlabel.toPlainText())

        # move elements
        src_view.selectionModel().select(
            gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        src_view.selectionModel().select(
            gene_index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.widget.move_button.click()

        # test in the dest view
        parent_index = dest_view.model().index(0, 0)
        gene_index = dest_view.model().index(0, 0, parent_index)
        gene_index1 = dest_view.model().index(1, 0, parent_index)

        # no description for parent
        dest_view.selectionModel().select(
            parent_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # no description for multiple genes
        dest_view.selectionModel().select(
            gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        dest_view.selectionModel().select(
            gene_index1, QItemSelectionModel.Select | QItemSelectionModel.Rows
        )
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # description for first gene
        dest_view.selectionModel().select(
            gene_index, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
        )
        self.assertIn(first_gene_info, self.widget.descriptionlabel.toPlainText())

    def test_extend_collapse_both_views(self):
        """
        Test whether the item extends in the other column when extended in one.
        """
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view

        # I do not know how to simulate click, so I just call the
        # function that is called on click
        index = src_view.model().index(0, 0)
        src_view.expand_in_other_view(index, is_expanded=True)
        tree_node = (
            src_view.model()
            .sourceModel()
            .node_from_index(src_view.model().mapToSource(index))
        )
        self.assertTrue(tree_node.expanded)

        # move one child left to check simultaneous expand
        self.move_first_child(src_view)
        index_left = src_view.model().index(0, 0)
        index_right = dest_view.model().index(0, 0)
        tree_node_left = (
            src_view.model()
            .sourceModel()
            .node_from_index(src_view.model().mapToSource(index_left))
        )
        tree_node_right = (
            dest_view.model()
            .sourceModel()
            .node_from_index(dest_view.model().mapToSource(index_right))
        )

        src_view.expand_in_other_view(index_left, is_expanded=True)
        self.assertTrue(tree_node_left.expanded)
        self.assertTrue(tree_node_right.expanded)
        self.assertTrue(dest_view.isExpanded(index_right))

        # change back to not expanded
        src_view.expand_in_other_view(index_left, is_expanded=False)
        self.assertFalse(tree_node_left.expanded)
        self.assertFalse(tree_node_right.expanded)
        self.assertFalse(dest_view.isExpanded(index_right))

        dest_view.expand_in_other_view(index_right, is_expanded=True)
        self.assertTrue(tree_node_left.expanded)
        self.assertTrue(tree_node_right.expanded)
        self.assertTrue(src_view.isExpanded(index_left))

        # change back to not expanded
        dest_view.expand_in_other_view(index_right, is_expanded=False)
        self.assertFalse(tree_node_left.expanded)
        self.assertFalse(tree_node_right.expanded)
        self.assertFalse(dest_view.isExpanded(index_left))


if __name__ == '__main__':
    unittest.main()



# pylint: disable=too-many-ancestors
from enum import IntEnum
from types import SimpleNamespace
from typing import Dict, Tuple, Optional
from itertools import chain

import numpy as np
import pyqtgraph as pg

from AnyQt.QtGui import QColor
from AnyQt.QtCore import Qt, QRectF, QObject

from Orange.data import Table, Domain, DiscreteVariable, ContinuousVariable
from Orange.widgets import gui, report
from Orange.data.util import array_equal
from Orange.data.filter import Values, FilterString
from Orange.widgets.widget import Msg, Input
from Orange.widgets.settings import Setting, ContextSetting, SettingProvider
from Orange.widgets.utils.sql import check_sql_input
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.colorpalettes import LimitedDiscretePalette
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.utils.widget import OWDataProjectionWidget
from Orange.widgets.visualize.owscatterplotgraph import OWScatterPlotBase

from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID
from orangecontrib.bioinformatics.annotation.annotate_samples import (
    PFUN_BINOMIAL,
    SCORING_LOG_FDR,
    SCORING_EXP_RATIO,
    PFUN_HYPERGEOMETRIC,
    SCORING_MARKERS_SUM,
    AnnotateSamplesMeta,
)
from orangecontrib.bioinformatics.annotation.annotate_projection import (
    annotate_projection,
    cluster_additional_points,
)

CELL_TYPE = "Cell Type"
ENTREZ_ID = "Entrez ID"


class Result(SimpleNamespace):
    scores: Optional[Table] = None
    clusters = None  # type: OWAnnotateProjection.Clusters


class Runner:
    @staticmethod
    def compute_scores(
        data: Table,
        genes: Table,
        p_threshold: float,
        p_value_fun: str,
        scoring: str,
        start: float,
        end: float,
        result: Result,
        state: TaskState,
    ):
        if not data or not genes:
            result.scores.z_vals = None
            result.scores.annotations = None
            result.scores.p_vals = None
            result.scores.table = None
        else:
            state.set_status("Computing scores...")
            weights = np.array([15, 75, 10]) * (end - start) / 100

            if not result.scores.z_vals:
                result.scores.z_vals = AnnotateSamplesMeta.mann_whitney_test(data)
                state.set_partial_result(("scores", result))
            state.set_progress_value(weights[0])
            if state.is_interruption_requested():
                return

            if not result.scores.annotations or not result.scores.p_vals:
                annot, p_vals = AnnotateSamplesMeta.assign_annotations(
                    result.scores.z_vals,
                    genes,
                    data,
                    p_value_fun=p_value_fun,
                    scoring=scoring,
                )
                result.scores.annotations = annot
                result.scores.p_vals = p_vals
                state.set_partial_result(("scores", result))
            state.set_progress_value(weights[1])
            if state.is_interruption_requested():
                return

            result.scores.table = AnnotateSamplesMeta.filter_annotations(
                result.scores.annotations, result.scores.p_vals, p_threshold=p_threshold
            )

        state.set_partial_result(("scores", result))

    @staticmethod
    def compute_clusters(embedding: Table, result: Result, state: TaskState):
        if not result.scores.table or not embedding:
            result.clusters.table = None
            result.clusters.groups = None
        else:
            state.set_status("Finding clusters...")
            kwargs = {}
            if result.clusters.epsilon is not None:
                kwargs["eps"] = result.clusters.epsilon
            clusters = annotate_projection(result.scores.table, embedding, **kwargs)
            result.clusters.table = clusters[0]
            result.clusters.groups = clusters[1]
            result.clusters.epsilon = clusters[2]
        state.set_partial_result(("clusters", result))

    @staticmethod
    def compute_secondary_clusters(embedding: Table, result: Result, state: TaskState):
        if not result.clusters.groups or not embedding:
            result.clusters.secondary_table = None
        else:
            state.set_status("Finding secondary clusters...")
            hulls = {k: v[2] for k, v in result.clusters.groups.items()}
            clusters = result.clusters.table
            domain = clusters and clusters.domain["Clusters"]
            table = cluster_additional_points(embedding, hulls, domain)
            result.clusters.secondary_table = table
        state.set_partial_result(("secondary_clusters", result))

    @classmethod
    def run(
        cls,
        data: Table,
        secondary_data: Table,
        attr_x: ContinuousVariable,
        attr_y: ContinuousVariable,
        genes: Table,
        p_threshold: float,
        p_value_fun: str,
        scoring: str,
        result: Result,
        state: TaskState,
    ):
        start, step, weights = 0, 0, np.array([80, 15, 5])

        def set_progress():
            nonlocal start
            nonlocal step
            start = int(start + weights[step])
            step += 1
            state.set_progress_value(start)
            return 0 if state.is_interruption_requested() else 1

        if not result.scores.table:
            end = start + weights[step]
            cls.compute_scores(
                data,
                genes,
                p_threshold,
                p_value_fun,
                scoring,
                start,
                end,
                result,
                state,
            )
        if not set_progress():
            return result

        if not result.clusters.table:
            embedding = data.transform(Domain([attr_x, attr_y]))
            cls.compute_clusters(embedding, result, state)
        if not set_progress():
            return result

        if not result.clusters.secondary_table and secondary_data:
            embedding = secondary_data.transform(Domain([attr_x, attr_y]))
            cls.compute_secondary_clusters(embedding, result, state)
        set_progress()
        return result


class CenteredTextItem(pg.TextItem):
    def __init__(self, view_box, x, y, text, tooltip):
        bg_color = QColor(Qt.white)
        bg_color.setAlpha(200)
        color = QColor(Qt.black)
        super().__init__(text, pg.mkColor(color), fill=pg.mkBrush(bg_color))
        self._x = x
        self._y = y
        self._view_box = view_box
        self._view_box.sigStateChanged.connect(self.center)
        self.textItem.setToolTip(tooltip)
        self.setPos(x, y)
        self.center()

    def center(self):
        br = self.boundingRect()
        dx = br.width() / 2 * self._view_box.viewPixelSize()[0]
        dy = br.height() / 2 * self._view_box.viewPixelSize()[1]
        self.setPos(self._x - dx, self._y + dy)


class EventDelegate(QObject):
    def eventFilter(self, *_):
        return False


class OWAnnotateProjectionGraph(OWScatterPlotBase):
    show_cluster_hull = Setting(True)
    n_cluster_labels = Setting(1)
    show_ref_data = Setting(True)

    def __init__(self, scatter_widget, parent):
        super().__init__(scatter_widget, parent)
        self.cluster_hulls_items = []
        self.cluster_labels_items = []
        self.ref_scatterplot_item = None
        self._tooltip_delegate = EventDelegate()  # remove points tooltip

    def clear(self):
        super().clear()
        self.cluster_hulls_items.clear()
        self.cluster_labels_items.clear()
        self.ref_scatterplot_item = None

    def reset_view(self):
        x, y = [self.get_coordinates()[0]], [self.get_coordinates()[1]]
        if x[0] is None or y[0] is None:
            return

        hulls = self.master.get_cluster_hulls()
        if hulls is not None:
            x.extend([hull[:, 0] for hull, _ in hulls])
            y.extend([hull[:, 1] for hull, _ in hulls])

        points = self.master.get_coordinates_reference_data()
        if points is not None:
            x.append(points[0])
            y.append(points[1])

        x, y = np.hstack(x), np.hstack(y)
        min_x, max_x, min_y, max_y = np.min(x), np.max(x), np.min(y), np.max(y)
        rect = QRectF(min_x, min_y, max_x - min_x or 1, max_y - min_y or 1)
        self.view_box.setRange(rect, padding=0.025)

    def update_coordinates(self):
        self.update_reference_coordinates()
        super().update_coordinates()
        self.update_clusters()
        self.view_box.setAspectLocked(True, 1)
        self.update_reference_item()
        self.reset_view()

    def update_clusters(self):
        self._update_cluster_hull()
        self._update_cluster_labels()

    def _update_cluster_hull(self):
        for item in self.cluster_hulls_items:
            self.plot_widget.removeItem(item)
        if not self.show_cluster_hull:
            return
        hulls = self.master.get_cluster_hulls()
        if hulls is None:
            return
        for hull, color in hulls:
            pen = pg.mkPen(color=QColor(*color), style=Qt.DashLine, width=3)
            item = pg.PlotCurveItem(x=hull[:, 0], y=hull[:, 1], pen=pen, antialias=True)
            self.plot_widget.addItem(item)
            self.cluster_hulls_items.append(item)

    def _update_cluster_labels(self):
        for item in self.cluster_labels_items:
            self.plot_widget.removeItem(item)
        if not self.n_cluster_labels:
            return
        labels = self.master.get_cluster_labels()
        if labels is None:
            return
        for label_per, (x, y), _ in labels:
            text = "\n".join([label for label, _ in label_per[: self.n_cluster_labels]])
            ttip = "\n".join([f"{round(p * 100)}%  {label}" for label, p in label_per])
            item = CenteredTextItem(self.view_box, x, y, text, ttip)
            self.plot_widget.addItem(item)
            self.cluster_labels_items.append(item)

    def update_reference_coordinates(self):
        points = self.master.get_coordinates_reference_data()
        if points is None:
            return
        if self.ref_scatterplot_item is None:
            color = pg.mkColor(200, 200, 200)
            pen, brush = pg.mkPen(color=color), pg.mkBrush(color=color)
            size = OWScatterPlotBase.MinShapeSize + 3
            self.ref_scatterplot_item = pg.ScatterPlotItem(
                x=points[0], y=points[1], pen=pen, brush=brush, size=size
            )
            self.plot_widget.addItem(self.ref_scatterplot_item)
        else:
            self.ref_scatterplot_item.setData(x=points[0], y=points[1])

    def update_reference_item(self):
        if self.ref_scatterplot_item is not None:
            self.ref_scatterplot_item.setVisible(self.show_ref_data)
        elif self.scatterplot_item is not None:
            self.scatterplot_item.setVisible(self.show_ref_data)


class ScoringMethod(IntEnum):
    ExpRatio, MarkersSum, LogFdr = range(3)

    @staticmethod
    def values():
        return [SCORING_LOG_FDR, SCORING_MARKERS_SUM, SCORING_EXP_RATIO]

    @staticmethod
    def items():
        return ["-Log(FDR)", "Marker expression", "Marker expression (%)"]


class StatisticalTest(IntEnum):
    Binomial, Hypergeometric = range(2)

    @staticmethod
    def values():
        return [PFUN_BINOMIAL, PFUN_HYPERGEOMETRIC]

    @staticmethod
    def items():
        return ["Binomial", "Hypergeometric"]


class OWAnnotateProjection(OWDataProjectionWidget, ConcurrentWidgetMixin):
    name = "Annotator"
    description = "Annotates projection clusters."
    icon = "icons/OWAnnotateProjection.svg"
    priority = 140
    keywords = ["annotate"]

    GRAPH_CLASS = OWAnnotateProjectionGraph
    graph = SettingProvider(OWAnnotateProjectionGraph)

    attr_x = ContextSetting(None)
    attr_y = ContextSetting(None)
    scoring_method = Setting(ScoringMethod.ExpRatio)
    statistical_test = Setting(StatisticalTest.Binomial)
    p_threshold = Setting(0.05)
    use_user_epsilon = Setting(False)
    epsilon = Setting(0)
    color_by_cluster = Setting(False)

    class Scores(SimpleNamespace):
        z_vals: Optional[Table] = None
        annotations: Optional[Table] = None
        p_vals: Optional[Table] = None
        table: Optional[Table] = None

    class Clusters(SimpleNamespace):
        table: Optional[Table] = None
        groups: Optional[Dict[str, Tuple]] = None
        epsilon: Optional[float] = None
        secondary_table: Optional[Table] = None

    class Inputs:
        data = Input("Reference Data", Table, default=True)
        secondary_data = Input("Secondary Data", Table)
        genes = Input("Genes", Table)

    class Information(OWDataProjectionWidget.Information):
        modified = Msg(
            "The parameter settings have been changed. Press "
            "\"Start\" to rerun with the new settings."
        )

    class Warning(OWDataProjectionWidget.Warning):
        no_genes = Msg("Missing Genes table on input.")
        missing_compute_value = Msg(
            "{} or {} is missing in Secondary data and"
            " cannot be computed from Reference data."
        )
        missing_tax_id = Msg(
            f"'{TAX_ID}' is missing in Reference data. " f"Try using 'Genes' widget."
        )
        missing_entrez_id = Msg("'Entred ID' is missing in Genes table.")
        missing_cell_type = Msg("'Cell Type' is missing in Genes table.")
        missing_tax_id_genes = Msg(
            f"'{TAX_ID}' is missing in Genes table. "
            f"Try using 'Marker Genes' widget."
        )
        different_tax_id = Msg(
            f"Data and Genes appear to describe different "
            f"organisms (mismatching {TAX_ID})."
        )
        same_axis_features = Msg(
            "Selected features for Axis x and " "Axis y should differ."
        )

    class Error(OWDataProjectionWidget.Error):
        no_reference_data = Msg("Missing Reference data on input.")
        no_continuous_vars = Msg("Data has no continuous variables.")
        not_enough_inst = Msg("Not enough instances in Reference data.")
        proj_error = Msg("An error occurred while annotating data.\n{}")
        sec_proj_error = Msg(
            "An error occurred while projecting " "Secondary data.\n{}"
        )

    def __init__(self):
        OWDataProjectionWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)
        # inputs
        self.reference_data: Optional[Table] = None
        self.secondary_data: Optional[Table] = None
        self.genes: Optional[Table] = None
        # annotations
        self.scores: OWAnnotateProjection.Scores = None
        self.clusters: OWAnnotateProjection.Clusters = None
        self.__invalidate_scores()
        self.__invalidate_clusters()

    # GUI
    def _add_controls(self):
        self.__add_annotation_controls()
        super()._add_controls()
        self.gui.add_control(
            self._effects_box,
            gui.hSlider,
            "Cluster labels:",
            master=self.graph,
            value="n_cluster_labels",
            minValue=0,
            maxValue=3,
            step=1,
            createLabel=False,
            callback=self.graph.update_clusters,
        )
        self._plot_box.children()[1].hide()  # Hide 'Show color regions'
        gui.checkBox(
            self._plot_box,
            self.graph,
            "show_cluster_hull",
            "Show cluster hull",
            callback=self.graph.update_clusters,
        )
        gui.checkBox(
            self._plot_box,
            self,
            "color_by_cluster",
            "Color points by cluster",
            callback=self.__color_by_cluster_changed,
        )
        gui.checkBox(
            self._plot_box,
            self.graph,
            "show_ref_data",
            "Show reference data",
            callback=self.__show_ref_data_changed,
        )

    def __add_annotation_controls(self):
        common_options = {
            'labelWidth': 100,
            'orientation': Qt.Horizontal,
            'sendSelectedValue': True,
            'contentsLength': 14,
        }
        box = gui.vBox(self.controlArea, box='Axes')
        ord = (DomainModel.METAS, DomainModel.ATTRIBUTES, DomainModel.CLASSES)
        mod = DomainModel(ord, valid_types=ContinuousVariable)
        gui.comboBox(
            box,
            self,
            "attr_x",
            label="Axis x:",
            model=mod,
            callback=self.__axis_attr_changed,
            **common_options,
        )
        gui.comboBox(
            box,
            self,
            "attr_y",
            label="Axis y:",
            model=mod,
            callback=self.__axis_attr_changed,
            **common_options,
        )

        box = gui.vBox(self.controlArea, box='Annotation')
        gui.comboBox(
            box,
            self,
            "scoring_method",
            label="Scoring method:",
            items=ScoringMethod.items(),
            orientation=Qt.Horizontal,
            contentsLength=13,
            labelWidth=100,
            callback=self.__scoring_combo_changed,
        )
        gui.comboBox(
            box,
            self,
            "statistical_test",
            label="Statistical test:",
            items=StatisticalTest.items(),
            orientation=Qt.Horizontal,
            labelWidth=100,
            callback=self.__scoring_combo_changed,
        )
        gui.doubleSpin(
            box,
            self,
            "p_threshold",
            0,
            1,
            0.01,
            label="FDR threshold:",
            callback=self.__p_threshold_changed,
        )
        hbox = gui.hBox(box)
        gui.checkBox(
            hbox,
            self,
            "use_user_epsilon",
            "ε for DBSCAN:",
            callback=self.__epsilon_check_changed,
        )
        self.epsilon_spin = gui.doubleSpin(
            hbox, self, "epsilon", 0, 10, 0.1, callback=self.__epsilon_changed
        )
        self.run_button = gui.button(box, self, "Start", self._toggle_run)

    @property
    def effective_variables(self):
        attr_x, attr_y = self.attr_x, self.attr_y
        if attr_x and attr_y:
            if attr_x == attr_y:
                return [attr_x]
            return [attr_x, attr_y]
        return []

    @property
    def effective_data(self):
        return self.reference_data.transform(Domain(self.effective_variables))

    @property
    def can_annotate(self):
        return (
            self.reference_data
            and self.genes
            and TAX_ID in self.reference_data.attributes
            and ENTREZ_ID in self.genes.domain
            and CELL_TYPE in self.genes.domain
            and TAX_ID in self.genes.attributes
            and self.reference_data.attributes[TAX_ID] == self.genes.attributes[TAX_ID]
            and self.attr_x is not None
            and self.attr_y is not None
            and self.attr_x is not self.attr_y
        )

    def __color_by_cluster_changed(self):
        self.controls.attr_color.setEnabled(not self.color_by_cluster)
        self.graph.update_colors()

    def __show_ref_data_changed(self):
        self.graph.update_reference_item()

    def __axis_attr_changed(self):
        self.__parameter_changed()
        self.setup_plot()

    def __scoring_combo_changed(self):
        self.__invalidate_scores_annotations()
        self.__parameter_changed()

    def __p_threshold_changed(self):
        self.__invalidate_scores_table()
        self.__parameter_changed()

    def __epsilon_changed(self):
        self.__parameter_changed()

    def __parameter_changed(self):
        self.__invalidate_clusters()
        self.Information.modified()
        self.Error.proj_error.clear()

    def __epsilon_check_changed(self):
        self.enable_epsilon_spin()
        if not self.use_user_epsilon:
            self.__epsilon_changed()

    def __invalidate_scores(self):
        self.scores = self.Scores(
            z_vals=None, annotations=None, p_vals=None, scores=None
        )

    def __invalidate_scores_table(self):
        self.scores.table = None

    def __invalidate_scores_annotations(self):
        self.scores.annotations = None
        self.scores.p_vals = None
        self.__invalidate_scores_table()

    def __invalidate_clusters(self):
        epsilon = self.epsilon if self.use_user_epsilon else None
        self.clusters = self.Clusters(
            table=None, groups=None, epsilon=epsilon, secondary_table=None
        )

    def __invalidate_clusters_secondary_table(self):
        self.clusters.secondary_table = None

    def _toggle_run(self):
        if self.task is not None:
            self.cancel()
            self.run_button.setText("Resume")
            self.commit.deferred()
        else:
            self._run()

    def _run(self):
        self.Information.modified.clear()
        self.graph.update_clusters()  # Remove cluster hulls and labels
        if not self.can_annotate:
            return

        self.run_button.setText("Stop")
        result = Result(scores=self.scores, clusters=self.clusters)
        self.start(
            Runner.run,
            self.reference_data,
            self.secondary_data,
            self.attr_x,
            self.attr_y,
            self.genes,
            self.p_threshold,
            StatisticalTest.values()[self.statistical_test],
            ScoringMethod.values()[self.scoring_method],
            result,
        )

    def on_partial_result(self, value: Tuple[str, Result]):
        which, result = value
        if which == "scores":
            self.scores = result.scores
        elif which == "clusters":
            self.clusters = result.clusters
            if result.clusters.epsilon is not None:
                self.epsilon = result.clusters.epsilon
            self.graph.update_clusters()
            self.graph.update_colors()
            self.graph.reset_view()
        elif which == "secondary_clusters":
            self.clusters.secondary_table = result.clusters.secondary_table
            self.graph.update_colors()

    def on_done(self, result: Result):
        self.scores = result.scores
        self.clusters = result.clusters
        if result.clusters.epsilon is not None:
            self.epsilon = result.clusters.epsilon
        self.run_button.setText("Start")
        self.commit.deferred()

    def on_exception(self, ex: Exception):
        self.Error.proj_error(ex)
        self.run_button.setText("Start")

    @Inputs.data
    @check_sql_input
    def set_data(self, data):
        attr_x, attr_y = self.attr_x, self.attr_y
        data_existed = self.reference_data is not None
        effective_data = self.effective_data if data_existed else None
        super().set_data(data)
        self.reference_data = self.data
        if not (
            data_existed
            and self.reference_data is not None
            and array_equal(effective_data.X, self.effective_data.X)
        ):
            self.clear()
            self.__invalidate_scores()
            self.__invalidate_clusters()
        elif attr_x is not self.attr_x or attr_y is not self.attr_y:
            self.clear()
            self.__invalidate_clusters()

    def check_data(self):
        self.Warning.missing_tax_id.clear()
        self.Error.no_continuous_vars.clear()
        self.Error.not_enough_inst.clear()
        if self.data:
            if len(self.data) < 2:
                self.Error.not_enough_inst()
                self.data = None
            elif not self.data.domain.has_continuous_attributes(True, True):
                self.Error.no_continuous_vars()
                self.data = None
            elif TAX_ID not in self.data.attributes:
                self.Warning.missing_tax_id()

    def init_attr_values(self):
        super().init_attr_values()
        domain = self.data.domain if self.data and len(self.data) else None
        model = self.controls.attr_x.model()
        model.set_domain(domain)
        self.attr_x = model[0] if model else None
        self.attr_y = model[1] if len(model) >= 2 else self.attr_x

    def enable_controls(self):
        super().enable_controls()
        self.enable_epsilon_spin()
        self.controls.attr_color.setEnabled(not self.color_by_cluster)

    def enable_epsilon_spin(self):
        self.epsilon_spin.setEnabled(self.use_user_epsilon)

    @Inputs.secondary_data
    @check_sql_input
    def set_secondary_data(self, data):
        self.secondary_data = data
        self.clear()
        self.__invalidate_clusters_secondary_table()

    @Inputs.genes
    def set_genes(self, genes: Optional[Table]):
        self.genes = genes
        self.__invalidate_scores_annotations()
        self.__invalidate_clusters()
        self.check_genes()
        self.cancel()

    def check_genes(self):
        self.Warning.missing_tax_id_genes.clear()
        self.Warning.missing_entrez_id.clear()
        self.Warning.missing_cell_type.clear()
        if self.genes:
            if ENTREZ_ID not in self.genes.domain:
                self.Warning.missing_entrez_id()
            elif CELL_TYPE not in self.genes.domain:
                self.Warning.missing_cell_type()
            elif TAX_ID not in self.genes.attributes:
                self.Warning.missing_tax_id_genes()

    def handleNewSignals(self):
        self.check_inputs()
        super().handleNewSignals()
        self._run()

    def check_inputs(self):
        self.Warning.no_genes.clear()
        self.Warning.different_tax_id.clear()
        self.Error.no_reference_data.clear()
        self.Error.proj_error.clear()
        if self.reference_data and not self.genes:
            self.Warning.no_genes()
        elif (
            self.reference_data
            and self.genes
            and TAX_ID in self.reference_data.attributes
            and TAX_ID in self.genes.attributes
            and (
                self.genes.attributes[TAX_ID] != self.reference_data.attributes[TAX_ID]
            )
        ):
            self.Warning.different_tax_id()
        elif self.secondary_data and not self.reference_data:
            self.Error.no_reference_data()

        self.data = self.reference_data and self.secondary_data or self.reference_data
        self.init_attr_values_secondary()

    def init_attr_values_secondary(self):
        saved_attrs = {
            "attr_color": self.attr_color,
            "attr_shape": self.attr_shape,
            "attr_size": self.attr_size,
            "attr_label": self.attr_label,
        }
        domain = self.data.domain if self.data and len(self.data) else None
        for name, attr in saved_attrs.items():
            model = getattr(self.controls, name).model()
            model.set_domain(domain)
            if domain is not None and attr and attr.name in domain:
                setattr(self, name, domain[attr.name])
            else:
                setattr(self, name, None)

    def get_coordinates_reference_data(self):
        if not self.secondary_data or not self.reference_data:
            return None
        x_data = self.reference_data.get_column(self.attr_x)
        y_data = self.reference_data.get_column(self.attr_y)
        return np.vstack((x_data, y_data))

    def get_embedding(self):
        def no_embedding_attrs():
            domain = self.secondary_data.domain
            return (
                self.attr_x.name not in domain
                and self.attr_x.compute_value is None
                or self.attr_y.name not in domain
                and self.attr_y.compute_value is None
            )

        self.Warning.missing_compute_value.clear()
        self.Warning.same_axis_features.clear()
        self.Error.sec_proj_error.clear()
        if not self.data:
            self.valid_data = None
            return None

        if self.attr_x is self.attr_y and self.attr_x is not None:
            self.Warning.same_axis_features()

        if self.secondary_data and self.reference_data:
            if no_embedding_attrs():
                self.Warning.missing_compute_value(self.attr_x, self.attr_y)
            try:
                domain = self.reference_data.domain
                self.data = self.secondary_data.transform(domain)
            except Exception as ex:
                self.Error.sec_proj_error(ex)
                return None

        x_data = self.get_column(self.attr_x, filter_valid=False)
        y_data = self.get_column(self.attr_y, filter_valid=False)
        if x_data is None or y_data is None:
            return None

        self.valid_data = np.isfinite(x_data) & np.isfinite(y_data)
        return np.vstack((x_data, y_data)).T

    def get_color_data(self):
        table = (
            self.clusters.secondary_table
            if self.secondary_data
            else self.clusters.table
        )
        if not self.color_by_cluster or not table:
            return super().get_color_data()

        attr = table.domain["Clusters"]
        all_data = table.get_column(attr)
        if all_data.dtype == object and attr.is_primitive():
            all_data = all_data.astype(float)
        if self.valid_data is not None:
            all_data = all_data[self.valid_data]
        return all_data

    def get_color_labels(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().get_color_labels()
        return self.clusters.table.domain["Clusters"].values

    def is_continuous_color(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().is_continuous_color()
        return False

    def get_palette(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().get_palette()

        colors = self.clusters.table.domain["Clusters"].colors
        # the second option is to keep widget backward compatible (Orange < 3.25)
        return (
            self.clusters.table.domain["Clusters"].palette
            if hasattr(self.clusters.table.domain["Clusters"], "palette")
            else LimitedDiscretePalette(number_of_colors=len(colors))
        )

    def get_cluster_hulls(self):
        if not self.clusters.groups:
            return None
        var = self.clusters.table.domain["Clusters"]
        return [
            (hull, var.colors[var.values.index(key)])
            for key, (_, _, hull) in self.clusters.groups.items()
        ]

    def get_cluster_labels(self):
        if not self.clusters.groups:
            return None
        var = self.clusters.table.domain["Clusters"]
        return [
            (ann, lab, var.colors[var.values.index(key)])
            for key, (ann, lab, _) in self.clusters.groups.items()
        ]

    def send_data(self):
        if self.clusters.secondary_table:
            group_sel, graph = None, self.graph
            ref_proj_data, sec_proj_data = self._get_projection_data()
            data = Table.concatenate([ref_proj_data, sec_proj_data], axis=0)
            data = self.__add_data_source_column(data)

            if graph.selection is not None:
                ref_group_sel = np.zeros(len(ref_proj_data), dtype=int)
                sec_group_sel = np.zeros(len(sec_proj_data), dtype=int)
                sec_group_sel[self.valid_data] = graph.selection
                group_sel = np.hstack([ref_group_sel, sec_group_sel])

            selection = graph.get_selection() + len(ref_proj_data)
            self.Outputs.selected_data.send(
                self._get_selected_data(data, selection, group_sel)
            )

            # keeping backward compatible - versions of Orange < 3.25 fails on first call
            try:
                self.Outputs.annotated_data.send(
                    self._get_annotated_data(data, group_sel, graph.selection)
                )
            except TypeError:
                self.Outputs.annotated_data.send(
                    self._get_annotated_data(
                        data, selection, group_sel, graph.selection
                    )
                )
        else:
            super().send_data()

    def _get_projection_data(self):
        if not self.scores.table or not self.clusters.table:
            return self.data

        def get_augmented_data(table, scores_x, clusters_x):
            new_table = table.transform(domain)
            with new_table.unlocked(new_table.metas):
                if scores_x is not None:
                    new_table.metas[:, :n_sco] = scores_x
                new_table.metas[:, n_sco:n_clu] = clusters_x
            return new_table

        n_sco = self.scores.table.X.shape[1]
        n_clu = n_sco + self.clusters.table.X.shape[1]
        domain = self.reference_data.domain
        metas = chain(
            self.scores.table.domain.attributes,
            self.clusters.table.domain.attributes,
            domain.metas,
        )
        domain = Domain(domain.attributes, domain.class_vars, metas)
        if self.clusters.secondary_table:
            domain = self.__concatenate_domains(domain, self.secondary_data.domain)

        augmented_data = get_augmented_data(
            self.reference_data, self.scores.table.X, self.clusters.table.X
        )
        if self.clusters.secondary_table:
            sec_aug_data = get_augmented_data(
                self.secondary_data, None, self.clusters.secondary_table.X
            )
            return augmented_data, sec_aug_data
        return augmented_data

    @staticmethod
    def __concatenate_domains(ref_domain, sec_domain):
        attributes = list(ref_domain.attributes)
        class_vars = list(ref_domain.class_vars)
        metas = list(ref_domain.metas)
        for attr in sec_domain.attributes:
            if attr.name not in ref_domain:
                attributes.append(attr)
        for attr in sec_domain.class_vars:
            if attr.name not in ref_domain:
                class_vars.append(attr)
        for attr in sec_domain.metas:
            if attr.name not in ref_domain:
                metas.append(attr)
        return Domain(attributes, class_vars, metas)

    def __add_data_source_column(self, data):
        source_attr = DiscreteVariable(
            "Data Source", values=[self.reference_data.name, self.secondary_data.name]
        )
        domain = data.domain
        metas = chain(domain.metas, [source_attr])
        domain = Domain(domain.attributes, domain.class_vars, metas)
        data = data.transform(domain)
        with data.unlocked(data.metas):
            data.metas[: len(self.reference_data), -1] = 0
            data.metas[len(self.reference_data) :, -1] = 1
        return data

    def _get_send_report_caption(self):
        color_vr_name = (
            "Clusters"
            if self.color_by_cluster
            else self._get_caption_var_name(self.attr_color)
        )
        return report.render_items_vert(
            (
                ("Color", color_vr_name),
                ("Label", self._get_caption_var_name(self.attr_label)),
                ("Shape", self._get_caption_var_name(self.attr_shape)),
                ("Size", self._get_caption_var_name(self.attr_size)),
                (
                    "Jittering",
                    self.graph.jitter_size != 0
                    and "{} %".format(self.graph.jitter_size),
                ),
            )
        )

    def clear(self):
        super().clear()
        self.cancel()
        self._invalidated = True

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()


if __name__ == "__main__":
    from Orange.projection import PCA

    from orangecontrib.bioinformatics.utils import serverfiles

    data_path = "https://datasets.biolab.si/sc/aml-1k.tab.gz"
    table_data = Table(data_path)
    table_data.attributes[TAX_ID] = "9606"

    ref_data = table_data[::2]
    pca = PCA(n_components=2)
    pca_model = pca(ref_data)
    proj = pca_model(ref_data)
    new_dom = Domain(
        ref_data.domain.attributes,
        ref_data.domain.class_vars,
        chain(ref_data.domain.metas, proj.domain.attributes),
    )
    ref_data = ref_data.transform(new_dom)

    genes_path = serverfiles.localpath_download(
        "marker_genes", "panglao_gene_markers.tab"
    )
    filter_ = FilterString("Organism", FilterString.Equal, "Human")
    table_genes = Values([filter_])(Table(genes_path))
    table_genes.attributes[TAX_ID] = "9606"

    WidgetPreview(OWAnnotateProjection).run(
        set_data=ref_data, set_secondary_data=table_data[1:200:2], set_genes=table_genes
    )



""" Documentation script """
import textwrap


from orangecontrib.bioinformatics.geo.dataset import GDSInfo

gds_info = GDSInfo()
gds = gds_info["GDS10"]

print("ID:")
print(gds["dataset_id"])
print("Features: ")
print(gds["feature_count"])
print("Genes:")
print(gds["gene_count"])
print("Organism:")
print(gds["platform_organism"])
print("PubMed ID:")
print(gds["pubmed_id"])
print("Sample types:")

for sample_type in set([sinfo["type"] for sinfo in gds["subsets"]]):
    ss = [sinfo["description"] for sinfo in gds["subsets"] if sinfo["type"] == sample_type]
    print("  %s (%s)" % (sample_type, ", ".join(ss)))

print("")
print("Description:")
print("\n".join(textwrap.wrap(gds["description"], 70)))



import numpy as np
from scipy.stats import zscore, rankdata
from sklearn.preprocessing import quantile_transform

from Orange.data.table import Table
from Orange.preprocess.preprocess import Preprocess


class LogarithmicScale(Preprocess):
    def __call__(self, data) -> Table:
        _data = data.copy()
        with _data.unlocked(_data.X):
            _data.X = np.log2(data.X + 1)
        return _data


class ZScore(Preprocess):
    """
    Compute the z score.

    Detailed description: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.zscore.html
    """

    def __init__(self, axis=0):
        self.axis = axis

    def __call__(self, data) -> Table:
        _data = data.copy()
        with _data.unlocked(_data.X):
            _data.X = zscore(data.X, axis=self.axis)
            _data.X[np.isnan(_data.X)] = 0
        return _data


class QuantileTransform(Preprocess):
    """
    Transform features to follow a uniform or a normal distribution.

    Detailed description: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.quantile_transform.html
    """

    def __init__(self, axis=0, n_quantiles=1000, output_distribution='uniform'):
        self.axis = axis
        self.n_quantiles = n_quantiles
        self.output_distribution = output_distribution

    def __call__(self, data) -> Table:
        _data = data.copy()
        _data.X = quantile_transform(
            _data.X,
            n_quantiles=self.n_quantiles,
            output_distribution=self.output_distribution,
            copy=True,
            axis=self.axis,
        )
        return _data


class QuantileNormalization(Preprocess):
    """
    Quantile normalize a test distribution to a reference distribution
    of the same length by taking the average of each quantile across samples.

    Detailed description: https://en.wikipedia.org/wiki/Quantile_normalization
    """

    def __call__(self, data) -> Table:
        _data = data.copy()

        mean = np.mean(np.sort(_data.X, axis=1), axis=0)
        rank = rankdata(_data.X, method='average', axis=1) - 1

        rank_floor = rank.astype(int)
        rank_ceil = np.ceil(rank).astype(int)
        with _data.unlocked(_data.X):
            _data.X = (mean.take(rank_floor) + mean.take(rank_ceil)) / 2

        return _data



import itertools

from orangewidget.utils.signals import PartialSummary, summarize

from orangecontrib.bioinformatics.geneset import GeneSets


@summarize.register
def summarize_(gene_sets: GeneSets):  # pylint: disable=function-redefined
    n = len(gene_sets)
    if n == 0:
        details = 'empty gene sets'
    elif n <= 3:
        details = ', '.join(gs.name for gs in gene_sets)
    else:
        details = (
            ', '.join(gs.name for gs in itertools.islice(gene_sets, 3))
            + f' and {n - 3} others'
        )

    return PartialSummary(n, details)



"""

"""
from __future__ import absolute_import

from . import conf

REST_API = "http://rest.kegg.jp/"


def slumber_service():
    """
    Return a rest based service using `slumber` package
    """
    import slumber

    if not hasattr(slumber_service, "_cached"):

        class DecodeSerializer(slumber.serialize.BaseSerializer):
            key = "decode"
            content_types = ["text/plain"]

            def loads(self, data):
                return data

        # for python 2/3 compatibility
        serializer = slumber.serialize.Serializer(default="decode", serializers=[DecodeSerializer()])
        slumber_service._cached = slumber.API(REST_API, serializer=serializer)
    return slumber_service._cached


default_service = slumber_service

web_service = slumber_service



import unittest

import six

from orangecontrib.bioinformatics.kegg import pathway, databases


class TestGenome(unittest.TestCase):
    def test_genome(self):
        genome = databases.Genome()
        entry_keys = list(genome.keys())

        for key in entry_keys[:3] + entry_keys[-3:]:
            self.assertTrue(key in genome)
            self.assertTrue(key in genome)
            entry = genome[key]
            self.assertEqual(entry.entry_key, key)
            self.assertIsInstance(entry, genome.ENTRY_TYPE)
            self.assertIsInstance(entry.name, six.string_types)
            self.assertIsInstance(entry.taxid, six.string_types)

        search_result = genome.search("homo sapiens")[0]
        org_code, name = search_result.split(';')
        self.assertEqual("hsa", org_code.strip())
        self.assertEqual("Homo sapiens (human)", name.strip())
        entry = genome['hsa']
        self.assertEqual(entry.taxid, "9606")


class TestGenes(unittest.TestCase):
    def _tester(self, org):
        genes = databases.Genes(org)
        keys = list(genes.keys())
        keys = keys[:3] + keys[-3:]
        all_entries = []
        for gene in keys:
            self.assertTrue(gene in genes)
            entry = genes[gene]
            self.assertEqual(entry.entry_key, genes.get(gene).entry_key, "__getitem__ and get return different result")

            self.assertTrue(gene.endswith(entry.entry_key))

            self.assertIsInstance(entry, genes.ENTRY_TYPE)
            self.assertIsInstance(entry.aliases(), list)

            self.assertTrue(all(isinstance(a, six.string_types) for a in entry.aliases()))
            all_entries.append(entry)

        self.assertSequenceEqual(
            [(e.name, e.entry_key) for e in all_entries],
            [(e.name, e.entry_key) for e in genes.batch_get(keys)],
            "batch_get returns different result",
        )

    def test_hsa(self):
        self._tester("hsa")

    def test_sce(self):
        self._tester("sce")

    def test_ddi(self):
        self._tester("ddi")


class TestPathways(unittest.TestCase):
    def _tester(self, path_id):
        pathways = databases.Pathway()

        entry = pathways[path_id]
        self.assertTrue(path_id.endswith(entry.entry_key))
        self.assertIsInstance(entry, pathways.ENTRY_TYPE)

        genes = entry.gene or []

        path = pathway.Pathway(path_id)
        self.assertEqual(sorted(genes), sorted(path.genes()))

    def test(self):
        self._tester("path:hsa05130")


class TestUtils(unittest.TestCase):
    def test_batch_iter(self):
        iter = range(25)
        expected = [list(range(10)), list(range(10, 20)), list(range(20, 25))]
        for exp, batch in zip(expected, databases.batch_iter(iter, 10)):
            self.assertEqual(exp, batch)


if __name__ == '__main__':
    unittest.main()



"""  Gene Ontology module """
import os
import re
import sys
import tarfile
import warnings
from collections import namedtuple, defaultdict

import six

from orangecontrib.bioinformatics.ncbi import taxonomy
from orangecontrib.bioinformatics.utils import statistics, serverfiles, progress_bar_milestones
from orangecontrib.bioinformatics.go.config import DOMAIN, FILENAME_ONTOLOGY, FILENAME_ANNOTATION

intern = sys.intern
default_database_path = os.path.join(serverfiles.localpath(), DOMAIN)

_CVS_REVISION_RE = re.compile(r"^(rev)?(\d+\.\d+)+$")

evidence_types = {
    # Experimental
    'EXP': 'Inferred from Experiment',
    'IDA': 'Inferred from Direct Assay',
    'IPI': 'Inferred from Physical Interaction',  # [with <database:protein_name>]',
    'IMP': 'Inferred from Mutant Phenotype',
    'IGI': 'Inferred from Genetic Interaction',  # [with <database:gene_symbol[allele_symbol]>]',
    'IEP': 'Inferred from Expression Pattern',
    # Computational Analysis Evidence Codes
    'ISS': 'Inferred from Sequence Similarity',  # [with <database:sequence_id>] ',
    'ISA': 'Inferred from Sequence Alignment',
    'ISO': 'Inferred from Sequence Orthology',
    'ISM': 'Inferred from Sequence Model',
    'IGC': 'Inferred from Genomic Context',
    'RCA': 'Inferred from Reviewed Computational Analysis',
    # Author Statement Evidence Codes
    'TAS': 'Traceable author statement',
    'NAS': 'Non-traceable author statement',
    # Curatorial Statement Evidence Codes
    'IC': 'Inferred by curator',
    'ND': 'No biological data available',
    # Computationally-assigned Evidence Codes
    'IEA': 'Inferred from electronic annotation',  # [to <database:id>]',
    # Obsolete Evidence Codes
    'NR': 'Not Recorded(Obsolete)',
}

evidence_dict = defaultdict(int, [(e, 2 ** i) for i, e in enumerate(evidence_types.keys())])

evidence_types_ordered = [
    'EXP',
    'IDA',
    'IPI',
    'IMP',
    'IGI',
    'IEP',
    # Computational Analysis Evidence Codes
    'ISS',
    'ISA',
    'ISO',
    'ISM',
    'IGC',
    'RCA',
    # Author Statement Evidence Codes
    'TAS',
    'NAS',
    # Curatorial Statement Evidence Codes
    'IC',
    'ND',
    # Computationally-assigned Evidence Codes
    'IEA',
    # Obsolete Evidence Codes
    'NR',
]

multiplicity_set = {
    "alt_id",
    "is_a",
    "subset",
    "synonym",
    "related_synonym",
    "exact_synonym",
    "broad_synonym",
    "narrow_synonym",
    "xref_analog",
    "xref_unknown",
    "relationship",
}

multiple_tag_set = multiplicity_set

builtin_obo_objects = [
    """
[Typedef]
id: is_a
name: is_a
range: OBO:TERM_OR_TYPE
domain: OBO:TERM_OR_TYPE
definition: The basic subclassing relationship [OBO:defs]""",
    """[Typedef]
id: disjoint_from
name: disjoint_from
range: OBO:TERM
domain: OBO:TERM
definition: Indicates that two classes are disjoint [OBO:defs]""",
    """[Typedef]
id: instance_of
name: instance_of
range: OBO:TERM
domain: OBO:INSTANCE
definition: Indicates the type of an instance [OBO:defs]""",
    """[Typedef]
id: inverse_of
name: inverse_of
range: OBO:TYPE
domain: OBO:TYPE
definition: Indicates that one relationship type is the inverse of another [OBO:defs]""",
    """[Typedef]
id: union_of
name: union_of
range: OBO:TERM
domain: OBO:TERM
definition: Indicates that a term is the union of several others [OBO:defs]""",
    """[Typedef]
id: intersection_of
name: intersection_of
range: OBO:TERM
domain: OBO:TERM
definition: Indicates that a term is the intersection of several others [OBO:defs]""",
]


class OBOObject:
    """ Represents a generic OBO object (e.g. Term, Typedef, Instance, ...)
    """

    _INTERN_TAGS = ["id", "name", "namespace", "alt_id", "is_a"]

    def __init__(self, stanza=None, ontology=None):
        self.ontology = ontology
        self._lines = []
        self.values = {}
        self.related = set()
        self.related_to = set()
        if stanza:
            self.parse_stanza(stanza)

    def parse_stanza(self, stanza):
        intern_tags = set(self._INTERN_TAGS)
        for line in stanza.splitlines():
            if ":" not in line:
                continue
            tag, rest = line.split(":", 1)
            value, modifiers, comment = "", "", ""
            if "!" in rest:
                rest, comment = rest.split("!")
            if "{" in rest:
                value, modifiers = rest.split("{", 1)
                modifiers = modifiers.strip("}")
            else:
                value = rest
            tag = intern(tag)
            value = value.strip()
            comment = comment.strip()
            if tag in intern_tags:
                value, comment = intern(value), intern(comment)
            self._lines.append((tag, value, modifiers, comment))
            if tag in multiple_tag_set:
                self.values.setdefault(tag, []).append(value)
            else:
                self.values[tag] = value
        self.related = set(self.related_objects())
        self.__dict__.update(self.values)
        if "def" in self.__dict__:
            self.__dict__["def_"] = self.__dict__["def"]

    def related_objects(self):
        """Return a list of tuple pairs where the first element is relationship
        type_id and the second id of object to whom the relationship applies to.

        """
        # TODO: add other defined Typedef ids
        type_ids = [intern("is_a")]
        result = [(type_id, id) for type_id in type_ids for id in self.values.get(type_id, [])]
        result = result + [tuple(map(intern, r.split(None, 1))) for r in self.values.get("relationship", [])]
        return result

    def __repr__(self):
        """ Return a string representation of the object in OBO format
        """
        _repr = "[%s]\n" % type(self).__name__
        for tag, value, modifiers, comment in self._lines:
            _repr = _repr + tag + ": " + value
            if modifiers:
                _repr = _repr + "{ " + modifiers + " }"
            if comment:
                _repr = _repr + " ! " + comment
            _repr = _repr + "\n"
        return _repr

    def __str__(self):
        """ Return the OBO object id entry
        """
        return "%s: %s" % (self.id, self.name)

    def __iter__(self):
        """ Iterates over sub terms
        """
        for type_id, id in self.related_to:
            yield (type_id, self.ontology[id])


class Term(OBOObject):
    pass


class Typedef(OBOObject):
    pass


class Instance(OBOObject):
    pass


class Ontology:
    """
    :class:`Ontology` is the class representing a gene ontology.

    :param str filename:
        A filename of an .obo formated file.
    :param progress_callback:
        Optional `float -> None` function.


    Example
    --------
        >>> # Load the current ontology (downloading it if necessary)
        >>> ontology = Ontology()

        >>> term_ids = list(ontology)
        >>> term = ontology[term_ids[0]]

    """

    version = 1

    def __init__(self, filename=None, progress_callback=None):
        self.terms = {}
        self.typedefs = {}
        self.instances = {}
        self.slims_subset = set()
        self.alias_mapper = {}
        self.reverse_alias_mapper = defaultdict(set)
        self.header = ""

        if filename is not None:
            self.parse_file(filename, progress_callback)
        else:
            filename = serverfiles.localpath_download(DOMAIN, FILENAME_ONTOLOGY)
            self.parse_file(filename, progress_callback)

    @classmethod
    def load(cls, progress_callback=None):
        """
        A class method that tries to load the ontology file from
        default_database_path. It looks for a filename starting with
        'gene_ontology'. If not found it will download it.

        """
        filename = serverfiles.localpath_download(DOMAIN, FILENAME_ONTOLOGY)

        return cls(filename, progress_callback=progress_callback)

    Load = load

    def parse_file(self, file, progress_callback=None):
        """
        Parse the file. file can be a filename string or an open file like
        object. The optional progressCallback will be called with a single
        argument to report on the progress.
        """
        if isinstance(file, str):
            if os.path.isfile(file) and tarfile.is_tarfile(file):
                f = tarfile.open(file).extractfile("gene_ontology_edit.obo")
            elif os.path.isfile(file):
                f = open(file)
            elif os.path.isdir(file):
                f = open(os.path.join(file, "gene_ontology_edit.obo"))
            else:
                raise ValueError("Cannot open %r for parsing" % file)
        else:
            f = file

        data = [line.decode() if not isinstance(line, str) else line for line in f.readlines()]
        data = "".join([line for line in data if not line.startswith("!")])
        self.header = data[: data.index("[Term]")]
        c = re.compile(r"\[.+?\].*?\n\n", re.DOTALL)
        data = c.findall(data)

        milestones = progress_bar_milestones(len(data), 90)
        for i, block in enumerate(builtin_obo_objects + data):
            if block.startswith("[Term]"):
                term = Term(block, self)
                self.terms[term.id] = term
            elif block.startswith("[Typedef]"):
                typedef = Typedef(block, self)
                self.typedefs[typedef.id] = typedef
            elif block.startswith("[Instance]"):
                instance = Instance(block, self)
                self.instances[instance.id] = instance
            if progress_callback and i in milestones:
                progress_callback(90.0 * i / len(data))

        self.alias_mapper = {}
        self.reverse_alias_mapper = defaultdict(set)
        milestones = progress_bar_milestones(len(self.terms), 10)
        for i, (id, term) in enumerate(six.iteritems(self.terms)):
            for type_id, parent in term.related:
                self.terms[parent].related_to.add((type_id, id))
            try:
                self.alias_mapper.update([(alt_id, id) for alt_id in term.alt_id])
                self.reverse_alias_mapper[id].update(term.alt_id)
            except AttributeError:
                pass
            if progress_callback and i in milestones:
                progress_callback(90.0 + 10.0 * i / len(self.terms))

    def defined_slims_subsets(self):
        """
        Return a list of defined subsets in the ontology.

        :rtype: :class:`list` of :class:`str`

        """
        return [line.split()[1] for line in self.header.splitlines() if line.startswith("subsetdef:")]

    def named_slims_subset(self, subset):
        """
        Return all term IDs in a named `subset`.

        :param str subset: A string naming a subset in the ontology.
        :rtype: :class:`list` of :class:`str`

        .. seealso:: :func:`defined_slims_subsets`

        """
        return [id for id, term in self.terms.items() if subset in getattr(term, "subset", set())]

    def set_slims_subset(self, subset):
        """
        Set the `slims_subset` term subset to `subset`.

        :param set subset: A subset of GO term IDs.

        `subset` may also be a string, in which case the call is equivalent
        to ``ont.set_slims_subsets(ont.named_slims_subset(subset))``

        """
        if isinstance(subset, str):
            self.slims_subset = set(self.named_slims_subset(subset))
        else:
            self.slims_subset = set(subset)

    def slims_for_term(self, term):
        """
        Return a list of slim term IDs for `term`.

        This is a list of `most specific` slim terms to which `term` belongs.

        :param str term: Term ID.

        """
        queue = {term}
        visited = set()
        slims = set()
        while queue:
            term = queue.pop()
            visited.add(term)
            if term in self.slims_subset:
                slims.add(term)
            else:
                queue.update({tid for _, tid in self[term].related} - visited)
        return slims

    def extract_super_graph(self, terms):
        """
        Return all super terms of `terms` up to the most general one.

        :param list terms: A list of term IDs.

        """
        terms = [terms] if isinstance(terms, str) else terms
        visited = set()
        queue = set(terms)
        while queue:
            term = queue.pop()
            visited.add(term)
            queue.update({tid for _, tid in self[term].related} - visited)
        return visited

    def extract_sub_graph(self, terms):
        """
        Return all sub terms of `terms`.

        :param list terms: A list of term IDs.

        """
        terms = [terms] if type(terms) == str else terms
        visited = set()
        queue = set(terms)
        while queue:
            term = queue.pop()
            visited.add(term)
            queue.update({tid for _, tid in self[term].related_to} - visited)
        return visited

    def term_depth(self, term, cache_={}):
        """
        Return the minimum depth of a `term`.

        (length of the shortest path to this term from the top level term).

        """
        if term not in cache_:
            cache_[term] = min([self.term_depth(parent) + 1 for _, parent in self[term].related] or [1])
        return cache_[term]

    def __getitem__(self, termid):
        """
        Return a :class:`Term` object with `termid`.

        :param str term: An id of a 'Term' in the ontology.
        :rtype: :class:`Term`

        """
        if termid in self.terms:
            return self.terms[termid]
        elif termid in self.alias_mapper:
            return self.terms[self.alias_mapper[termid]]
        else:
            raise KeyError(termid)

    def __iter__(self):
        """
        Iterate over all term ids in ontology.
        """
        return iter(self.terms)

    def __len__(self):
        """
        Return number of terms in ontology.
        """
        return len(self.terms)

    def __contains__(self, termid):
        """
        Return `True` if a term with `termid` is present in the ontology.
        """
        return termid in self.terms or termid in self.alias_mapper


annotation_fields = ["tax_id", "gene_id", "go_id", "evidence", "qualifier", "go_term", "pubMed", "aspect"]


_AnnotationRecordBase = namedtuple("AnnotationRecord", annotation_fields)


class AnnotationRecord(_AnnotationRecordBase):
    """ An annotation record mapping a gene to a term.

    See ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/README for description
    if individual fields under <gene2go> section.

    """

    def __new__(cls, *args):
        if len(args) == 1 and isinstance(args[0], str):
            args = map(intern, args[0].split("\t"))
        return super(AnnotationRecord, cls).__new__(cls, *args)

    @classmethod
    def from_string(cls, string):
        """ Create an instance from a line in a annotations file format from serverfiles.
        """
        return AnnotationRecord._make(map(intern, string.strip().split('\t')))


class Annotations:
    """ :class:`Annotations` object holds the annotations.

    :param str organism:
        an organism specifier (e.g. ``'9606'``). Annotations for that organism will be loaded.

    :param ontology: :class:`Ontology` object for annotations
    :type ontology: :class:`Ontology`

    """

    def __init__(self, organism, ontology=None, progress_callback=None, filename=None):
        #: A dictionary mapping a gene (gene_id) to a set of all annotations of that gene.
        self.gene_annotations = defaultdict(list)

        #: A dictionary mapping a GO term id to a set of annotations that are directly annotated to that term
        self.term_anotations = defaultdict(list)

        self.all_annotations = defaultdict(list)

        self._gene_names = None
        self._gene_names_dict = None

        #: A list of all :class:`AnnotationRecords` instances.
        self.annotations = []
        self.header = ''
        self.taxid = organism

        self._ontology = ontology

        if filename is None:
            try:
                filename = serverfiles.localpath_download(
                    DOMAIN, FILENAME_ANNOTATION.format(organism), progress_callback=progress_callback
                )
            except FileNotFoundError:
                raise taxonomy.UnknownSpeciesIdentifier(organism)

        self._parse_file(filename)

    @property
    def ontology(self):
        return self._ontology

    @ontology.setter
    def ontology(self, ontology):
        """ Set the ontology to use in the annotations mapping.
        """
        self.all_annotations = defaultdict(list)
        self._ontology = ontology

    def _ensure_ontology(self):
        if self.ontology is None:
            self.ontology = Ontology()

    def _parse_file(self, file_path):

        with open(file_path, 'r') as anno_file:
            self.header = anno_file.readline()

            for line in anno_file.readlines():
                self.add_annotation(AnnotationRecord.from_string(line))

    def add_annotation(self, a):
        """ Add a single :class:`AnotationRecord` instance to this object.
        """
        if not isinstance(a, AnnotationRecord):
            a = AnnotationRecord(a)
        if not a.gene_id or not a.go_id or a.qualifier == 'NOT':
            return

        self.gene_annotations[a.gene_id].append(a)
        self.term_anotations[a.go_id].append(a)

        self.annotations.append(a)
        self.all_annotations = defaultdict(list)

    def get_genes_with_known_annotation(self, genes):
        """ Return only genes with known annotation

        :param genes: List of genes

        """
        return {gene for gene in genes if self.gene_annotations[gene]}

    def _collect_annotations(self, go_id, visited):
        """ Recursive function collects and caches all annotations for id
        """
        if go_id not in self.all_annotations and go_id not in visited:
            if go_id in self.ontology.reverse_alias_mapper:
                annotations = [
                    self.term_anotations.get(alt_id, []) for alt_id in self.ontology.reverse_alias_mapper[go_id]
                ] + [self.term_anotations[go_id]]
            else:
                annotations = [self.term_anotations[go_id]]  # annotations for this term alone
            visited.add(go_id)

            for type_id, child in self.ontology[go_id].related_to:
                aa = self._collect_annotations(child, visited)
                if type(aa) == set:
                    annotations.append(aa)  # if it was already reduced in get_all_annotations
                else:
                    annotations.extend(aa)
            self.all_annotations[go_id] = annotations
        return self.all_annotations[go_id]

    def get_annotations_by_go_id(self, go_id):
        """ Return a set of all annotations (instances of :obj:`AnnotationRecord`)
        for GO term `id` and all it's subterms.

        :param :class:`str` go_id: GO term id

        """
        self._ensure_ontology()
        id = self.ontology.alias_mapper.get(go_id, go_id)
        if id not in self.all_annotations or type(self.all_annotations[id]) == list:
            annot_set = set()
            for annots in self._collect_annotations(id, set()):
                annot_set.update(annots)
            self.all_annotations[id] = annot_set
        return self.all_annotations[id]

    def get_genes_by_go_term(self, go_id, evidence_codes=None):
        """ Return a list of genes annotated by specified `evidence_codes`
        to GO term 'id' and all it's subterms."

        :param str go_id: GO term id

        :param list-of-strings evidence_codes:
               List of evidence codes to consider when matching annotations to terms.

        """
        evidence_codes = set(evidence_codes or evidence_dict.keys())
        annotations = self.get_annotations_by_go_id(go_id)
        return list({int(ann.gene_id) for ann in annotations if ann.evidence in evidence_codes})

    def genes(self):
        return {ann.gene_id for ann in self.annotations}

    def get_enriched_terms(
        self,
        genes,
        reference=None,
        evidence_codes=None,
        slims_only=False,
        aspect=None,
        prob=statistics.Binomial(),
        use_fdr=True,
        progress_callback=None,
    ):
        """
        Return a dictionary of enriched terms, with tuples of
        (list_of_genes, p_value, reference_count) for items and term
        ids as keys. P-Values are FDR adjusted if use_fdr is True (default).

        :param genes: List of genes
        :param reference: List of genes (if None all genes included in the annotations will be used).
        :param evidence_codes:  List of evidence codes to consider.
        :param slims_only: If `True` return only slim terms.
        :param aspect: Which aspects to use. Use all by default;
                       one of Process (biological process),
                       Function (molecular function) or Component (cellular component)
        :param prob:
        :param use_fdr:
        :param progress_callback:
        """

        all_genes = set(genes)

        if aspect is None:
            aspects_set = {'Process', 'Component', 'Function'}
        elif isinstance(aspect, str):
            aspects_set = {aspect}
        else:
            aspects_set = aspect

        if reference is None:
            reference = self.genes()

        evidence_codes = set(evidence_codes or evidence_dict.keys())
        annotations = [
            ann
            for gene in genes
            for ann in self.gene_annotations[gene]
            if ann.evidence in evidence_codes and ann.aspect in aspects_set
        ]

        ref_annotations = {
            ann
            for gene in reference
            for ann in self.gene_annotations[gene]
            if ann.evidence in evidence_codes and ann.aspect in aspects_set
        }

        annotations_dict = defaultdict(set)
        for ann in annotations:
            annotations_dict[ann.go_id].add(ann)

        self._ensure_ontology()

        if slims_only and not self.ontology.slims_subset:
            warnings.warn("Unspecified slims subset in the ontology! " "Using 'goslim_generic' subset", UserWarning)
            self.ontology.set_slims_subset('goslim_generic')

        terms = annotations_dict.keys()
        filtered_terms = [term for term in terms if term in self.ontology]

        if len(terms) != len(filtered_terms):
            term_diff = set(terms) - set(filtered_terms)
            warnings.warn(
                "%s terms in the annotations were not found in the " "ontology." % ",".join(map(repr, term_diff)),
                UserWarning,
            )

        terms = self.ontology.extract_super_graph(filtered_terms)
        res = {}

        milestones = progress_bar_milestones(len(terms), 100)

        for i, term in enumerate(terms):
            if slims_only and term not in self.ontology.slims_subset:
                continue
            all_annotations = self.get_annotations_by_go_id(term).intersection(ref_annotations)
            all_annotated_genes = {ann.gene_id for ann in all_annotations}
            mapped_genes = all_genes.intersection(all_annotated_genes)

            if len(reference) > len(all_annotated_genes):
                mapped_reference_genes = reference.intersection(all_annotated_genes)
            else:
                mapped_reference_genes = all_annotated_genes.intersection(reference)

            res[term] = (
                [gene for gene in mapped_genes],
                prob.p_value(len(mapped_genes), len(reference), len(mapped_reference_genes), len(genes)),
                len(mapped_reference_genes),
            )

            if progress_callback and i in milestones:
                progress_callback(100.0 * i / len(terms))

        if use_fdr:
            res = sorted(res.items(), key=lambda x: x[1][1])
            res = {
                id: (genes, p, ref)
                for (id, (genes, _, ref)), p in zip(res, statistics.FDR([p for _, (_, p, _) in res]))
            }
        return res

    def get_annotated_terms(self, genes, direct_annotation_only=False, evidence_codes=None, progress_callback=None):
        """ Return all terms that are annotated by genes with evidence_codes.
        """

        genes = [genes] if type(genes) == str else genes
        genes = {gene for gene in genes}

        evidence_codes = set(evidence_codes or evidence_dict.keys())
        annotations = [ann for gene in genes for ann in self.gene_annotations[gene] if ann.evidence in evidence_codes]

        dd = defaultdict(set)
        for ann in annotations:
            dd[ann.go_id].add(ann.gene_id)

        if not direct_annotation_only:
            self._ensure_ontology()
            terms = dd.keys()
            filtered_terms = [term for term in terms if term in self.ontology]
            if len(terms) != len(filtered_terms):
                term_diff = set(terms) - set(filtered_terms)
                warnings.warn(
                    "%s terms in the annotations were not found in the " "ontology." % ",".join(map(repr, term_diff)),
                    UserWarning,
                )

            terms = self.ontology.extract_super_graph(filtered_terms)
            for i, term in enumerate(terms):
                term_annotations = self.get_annotations_by_go_id(term).intersection(annotations)
                dd[term].update([ann.gene_id for ann in term_annotations])
        return dict(dd)

    def __add__(self, iterable):
        """ Return a new Annotations object with combined annotations
        """
        return Annotations([a for a in self] + [a for a in iterable], ontology=self.ontology)

    def __iadd__(self, iterable):
        """ Add annotations to this instance
        """
        self.extend(iterable)
        return self

    def __contains__(self, item):
        return item in self.annotations

    def __iter__(self):
        """ Iterate over all AnnotationRecord objects in annotations
        """
        return iter(self.annotations)

    def __len__(self):
        """ Return the number of annotations
        """
        return len(self.annotations)

    def __getitem__(self, index):
        """ Return the i-th annotation record
        """
        return self.annotations[index]

    def __getslice__(self, *args):
        return self.annotations.__getslice__(*args)

    def add(self, line):
        """ Add one annotation
        """
        self.add_annotation(line)

    def append(self, line):
        """ Add one annotation
        """
        self.add_annotation(line)

    def extend(self, lines):
        """ Add multiple annotations
        """
        for line in lines:
            self.add_annotation(line)


def filter_by_p_value(terms, p_value=0.01):
    """ Filters the terms by the p-value. Assumes terms is a dict with
    the same structure as returned from get_enriched_terms.
    """
    return dict(filter(lambda x: x[1][1] <= p_value, terms.items()))


if __name__ == "__main__":
    pass
